"""Load and parse FIT files."""
import gzip

from fitparse import FitFile

CONVERSION_FACTOR = 180 / 2 ** 31
UNIVERSAL_FIELDS = {"time": lambda p: p["timestamp"], "dist": lambda p: p["distance"]}
NORMAL_FIELDS = {
    **UNIVERSAL_FIELDS,
    **{
        "lat": lambda p: p["position_lat"] * CONVERSION_FACTOR,
        "lon": lambda p: p["position_long"] * CONVERSION_FACTOR,
        "ele": lambda p: p["altitude"],
        "cadence": lambda p: p["cadence"] / 60,
        "heartrate": lambda p: p["heart_rate"] / 60,
        "speed": lambda p: p["speed"],
        "power": lambda p: p["power"],
    },
}

LENGTH_SWIM_FIELDS = {
    **UNIVERSAL_FIELDS,
    **{"speed": lambda p: p["speed"], "stroke": lambda p: p["swim_stroke"]},
}


def load_fit(filename):
    """Load and parse a FIT file."""
    if filename.suffix == ".gz":
        with gzip.open(filename) as f:
            fit = FitFile(f).messages
    else:
        fit = FitFile(str(filename)).messages
    fit = list(fit)
    for message in fit:
        point = message.get_values()
        if point.items() & {"sub_sport": "lap_swimming", "event": "length"}.items():
            return parse_fit(filename, fit, LENGTH_SWIM_FIELDS)
    return parse_fit(filename, fit, NORMAL_FIELDS)


def parse_fit(filename, fit, available_fields):
    """Extract the useful fields from a FIT file."""
    sport = "unknown"
    fields = {field: [] for field in available_fields}
    for message in fit:
        point = message.get_values()
        if message.mesg_type.name == "session":
            sport = point["sport"]

        elif message.mesg_type.name == "record":
            for field in available_fields:
                try:
                    value = available_fields[field](point)
                except Exception:
                    value = None
                fields[field].append(value)
    fields = {field: fields[field] for field in fields if set(fields[field]) != {None}}
    return (None, sport, fields)



# -*- coding: utf-8 -*-

# Form implementation generated from reading ui file 'activate/resources/ui/summary.ui'
#
# Created by: PyQt5 UI code generator 5.15.4
#
# WARNING: Any manual changes made to this file will be lost when pyuic5 is
# run again.  Do not edit this file unless you know what you are doing.


from PyQt5 import QtCore, QtGui, QtWidgets


class Ui_summary(object):
    def setupUi(self, summary):
        summary.setObjectName("summary")
        summary.resize(632, 254)
        self.summary_layout = QtWidgets.QHBoxLayout(summary)
        self.summary_layout.setObjectName("summary_layout")
        self.summary_options_layout = QtWidgets.QVBoxLayout()
        self.summary_options_layout.setObjectName("summary_options_layout")
        self.activity_types_list = CheckList(summary)
        self.activity_types_list.setMaximumSize(QtCore.QSize(300, 16777215))
        self.activity_types_list.setObjectName("activity_types_list")
        self.summary_options_layout.addWidget(self.activity_types_list)
        self.summary_period_layout = QtWidgets.QHBoxLayout()
        self.summary_period_layout.setObjectName("summary_period_layout")
        self.label = QtWidgets.QLabel(summary)
        self.label.setObjectName("label")
        self.summary_period_layout.addWidget(self.label)
        self.summary_period_box = QtWidgets.QComboBox(summary)
        self.summary_period_box.setMaximumSize(QtCore.QSize(300, 16777215))
        self.summary_period_box.setObjectName("summary_period_box")
        self.summary_period_box.addItem("")
        self.summary_period_box.addItem("")
        self.summary_period_box.addItem("")
        self.summary_period_box.addItem("")
        self.summary_period_layout.addWidget(self.summary_period_box)
        self.summary_options_layout.addLayout(self.summary_period_layout)
        self.summary_layout.addLayout(self.summary_options_layout)
        self.summary_tabs = QtWidgets.QTabWidget(summary)
        self.summary_tabs.setObjectName("summary_tabs")
        self.totals_tab = QtWidgets.QWidget()
        self.totals_tab.setObjectName("totals_tab")
        self.totals_tab_layout = QtWidgets.QFormLayout(self.totals_tab)
        self.totals_tab_layout.setObjectName("totals_tab_layout")
        self.distance_stat_label = QtWidgets.QLabel(self.totals_tab)
        font = QtGui.QFont()
        font.setPointSize(20)
        font.setBold(True)
        font.setWeight(75)
        self.distance_stat_label.setFont(font)
        self.distance_stat_label.setObjectName("distance_stat_label")
        self.totals_tab_layout.setWidget(
            0, QtWidgets.QFormLayout.LabelRole, self.distance_stat_label
        )
        self.total_distance_label = QtWidgets.QLabel(self.totals_tab)
        font = QtGui.QFont()
        font.setPointSize(12)
        self.total_distance_label.setFont(font)
        self.total_distance_label.setObjectName("total_distance_label")
        self.totals_tab_layout.setWidget(
            0, QtWidgets.QFormLayout.FieldRole, self.total_distance_label
        )
        self.time_stats_label = QtWidgets.QLabel(self.totals_tab)
        font = QtGui.QFont()
        font.setPointSize(20)
        font.setBold(True)
        font.setWeight(75)
        self.time_stats_label.setFont(font)
        self.time_stats_label.setObjectName("time_stats_label")
        self.totals_tab_layout.setWidget(
            1, QtWidgets.QFormLayout.LabelRole, self.time_stats_label
        )
        self.total_time_label = QtWidgets.QLabel(self.totals_tab)
        font = QtGui.QFont()
        font.setPointSize(12)
        self.total_time_label.setFont(font)
        self.total_time_label.setObjectName("total_time_label")
        self.totals_tab_layout.setWidget(
            1, QtWidgets.QFormLayout.FieldRole, self.total_time_label
        )
        self.activities_stat_label = QtWidgets.QLabel(self.totals_tab)
        font = QtGui.QFont()
        font.setPointSize(20)
        font.setBold(True)
        font.setWeight(75)
        self.activities_stat_label.setFont(font)
        self.activities_stat_label.setObjectName("activities_stat_label")
        self.totals_tab_layout.setWidget(
            2, QtWidgets.QFormLayout.LabelRole, self.activities_stat_label
        )
        self.total_activities_label = QtWidgets.QLabel(self.totals_tab)
        font = QtGui.QFont()
        font.setPointSize(12)
        self.total_activities_label.setFont(font)
        self.total_activities_label.setObjectName("total_activities_label")
        self.totals_tab_layout.setWidget(
            2, QtWidgets.QFormLayout.FieldRole, self.total_activities_label
        )
        self.climb_stats_label = QtWidgets.QLabel(self.totals_tab)
        font = QtGui.QFont()
        font.setPointSize(20)
        font.setBold(True)
        font.setWeight(75)
        self.climb_stats_label.setFont(font)
        self.climb_stats_label.setObjectName("climb_stats_label")
        self.totals_tab_layout.setWidget(
            3, QtWidgets.QFormLayout.LabelRole, self.climb_stats_label
        )
        self.total_climb_label = QtWidgets.QLabel(self.totals_tab)
        font = QtGui.QFont()
        font.setPointSize(12)
        self.total_climb_label.setFont(font)
        self.total_climb_label.setObjectName("total_climb_label")
        self.totals_tab_layout.setWidget(
            3, QtWidgets.QFormLayout.FieldRole, self.total_climb_label
        )
        self.summary_tabs.addTab(self.totals_tab, "")
        self.records_tab = QtWidgets.QWidget()
        self.records_tab.setObjectName("records_tab")
        self.verticalLayout_2 = QtWidgets.QVBoxLayout(self.records_tab)
        self.verticalLayout_2.setObjectName("verticalLayout_2")
        self.records_table = RecordsTable(self.records_tab)
        self.records_table.setEditTriggers(QtWidgets.QAbstractItemView.NoEditTriggers)
        self.records_table.setObjectName("records_table")
        self.records_table.setColumnCount(0)
        self.records_table.setRowCount(0)
        self.verticalLayout_2.addWidget(self.records_table)
        self.summary_tabs.addTab(self.records_tab, "")
        self.progression_tab = QtWidgets.QWidget()
        self.progression_tab.setObjectName("progression_tab")
        self.verticalLayout_5 = QtWidgets.QVBoxLayout(self.progression_tab)
        self.verticalLayout_5.setObjectName("verticalLayout_5")
        self.progression_chart = FullProgressionChart(self.progression_tab)
        self.progression_chart.setObjectName("progression_chart")
        self.verticalLayout_5.addWidget(self.progression_chart)
        self.summary_tabs.addTab(self.progression_tab, "")
        self.gallery_tab = QtWidgets.QWidget()
        self.gallery_tab.setObjectName("gallery_tab")
        self.verticalLayout_4 = QtWidgets.QVBoxLayout(self.gallery_tab)
        self.verticalLayout_4.setObjectName("verticalLayout_4")
        self.gallery_scroll_area = QtWidgets.QScrollArea(self.gallery_tab)
        self.gallery_scroll_area.setWidgetResizable(True)
        self.gallery_scroll_area.setObjectName("gallery_scroll_area")
        self.gallery = Gallery()
        self.gallery.setGeometry(QtCore.QRect(0, 0, 96, 26))
        self.gallery.setObjectName("gallery")
        self.gallery_scroll_area.setWidget(self.gallery)
        self.verticalLayout_4.addWidget(self.gallery_scroll_area)
        self.summary_tabs.addTab(self.gallery_tab, "")
        self.heatmap_tab = QtWidgets.QWidget()
        self.heatmap_tab.setObjectName("heatmap_tab")
        self.heatmap_layout = QtWidgets.QVBoxLayout(self.heatmap_tab)
        self.heatmap_layout.setObjectName("heatmap_layout")
        self.summary_tabs.addTab(self.heatmap_tab, "")
        self.eddington_tab = QtWidgets.QWidget()
        self.eddington_tab.setObjectName("eddington_tab")
        self.eddington_tab_layout = QtWidgets.QVBoxLayout(self.eddington_tab)
        self.eddington_tab_layout.setObjectName("eddington_tab_layout")
        self.eddington_number_form = QtWidgets.QFormLayout()
        self.eddington_number_form.setObjectName("eddington_number_form")
        self.eddington_stats_label = QtWidgets.QLabel(self.eddington_tab)
        font = QtGui.QFont()
        font.setPointSize(20)
        font.setBold(True)
        font.setWeight(75)
        self.eddington_stats_label.setFont(font)
        self.eddington_stats_label.setObjectName("eddington_stats_label")
        self.eddington_number_form.setWidget(
            0, QtWidgets.QFormLayout.LabelRole, self.eddington_stats_label
        )
        self.total_eddington_label = QtWidgets.QLabel(self.eddington_tab)
        font = QtGui.QFont()
        font.setPointSize(12)
        font.setBold(False)
        font.setWeight(50)
        self.total_eddington_label.setFont(font)
        self.total_eddington_label.setObjectName("total_eddington_label")
        self.eddington_number_form.setWidget(
            0, QtWidgets.QFormLayout.FieldRole, self.total_eddington_label
        )
        self.eddington_tab_layout.addLayout(self.eddington_number_form)
        self.eddington_chart_widget = QtChart.QChartView(self.eddington_tab)
        self.eddington_chart_widget.setObjectName("eddington_chart_widget")
        self.eddington_tab_layout.addWidget(self.eddington_chart_widget)
        self.summary_tabs.addTab(self.eddington_tab, "")
        self.summary_layout.addWidget(self.summary_tabs)

        self.retranslateUi(summary)
        self.summary_period_box.setCurrentIndex(1)
        self.summary_tabs.setCurrentIndex(0)
        self.progression_chart.y_changed.connect(summary.summary_tab_switch)
        self.activity_types_list.itemChanged["QListWidgetItem*"].connect(
            summary.summary_tab_switch
        )
        self.activity_types_list.itemDoubleClicked["QListWidgetItem*"].connect(
            summary.summary_tab_switch
        )
        self.summary_tabs.currentChanged["int"].connect(summary.summary_tab_switch)
        self.summary_period_box.currentIndexChanged["QString"].connect(
            summary.summary_period_changed
        )
        QtCore.QMetaObject.connectSlotsByName(summary)

    def retranslateUi(self, summary):
        _translate = QtCore.QCoreApplication.translate
        self.label.setText(_translate("summary", "Period"))
        self.summary_period_box.setItemText(0, _translate("summary", "All Time"))
        self.summary_period_box.setItemText(1, _translate("summary", "Year"))
        self.summary_period_box.setItemText(2, _translate("summary", "Month"))
        self.summary_period_box.setItemText(3, _translate("summary", "Week"))
        self.distance_stat_label.setText(_translate("summary", "Distance"))
        self.total_distance_label.setText(_translate("summary", "TextLabel"))
        self.time_stats_label.setText(_translate("summary", "Time"))
        self.total_time_label.setText(_translate("summary", "TextLabel"))
        self.activities_stat_label.setText(_translate("summary", "Activities"))
        self.total_activities_label.setText(_translate("summary", "TextLabel"))
        self.climb_stats_label.setText(_translate("summary", "Climb"))
        self.total_climb_label.setText(_translate("summary", "TextLabel"))
        self.summary_tabs.setTabText(
            self.summary_tabs.indexOf(self.totals_tab), _translate("summary", "Totals")
        )
        self.summary_tabs.setTabText(
            self.summary_tabs.indexOf(self.records_tab),
            _translate("summary", "Records"),
        )
        self.summary_tabs.setTabText(
            self.summary_tabs.indexOf(self.progression_tab),
            _translate("summary", "Progression"),
        )
        self.summary_tabs.setTabText(
            self.summary_tabs.indexOf(self.gallery_tab),
            _translate("summary", "Gallery"),
        )
        self.summary_tabs.setTabText(
            self.summary_tabs.indexOf(self.heatmap_tab),
            _translate("summary", "Heatmap"),
        )
        self.eddington_stats_label.setText(_translate("summary", "Eddington Number"))
        self.total_eddington_label.setText(_translate("summary", "TextLabel"))
        self.summary_tabs.setTabText(
            self.summary_tabs.indexOf(self.eddington_tab),
            _translate("summary", "Eddington"),
        )


from PyQt5 import QtChart
from activate.app.charts import FullProgressionChart
from activate.app.checklist import CheckList
from activate.app.photos import Gallery
from activate.app.tables import RecordsTable



"""Display an individual activity's data."""
import markdown
from PyQt5 import QtWidgets
from PyQt5.QtCore import pyqtSignal

from activate import activity_types, times
from activate.app import charts, photos
from activate.app.ui.activity_view import Ui_activity_view


class ActivityView(QtWidgets.QWidget, Ui_activity_view):
    """The statistics, graphs and map showing an activity."""

    closed = pyqtSignal()

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.setupUi(self)

    def setup(self, unit_system, map_widget):
        self.unit_system = unit_system
        self.updated = set()
        self.map_widget = map_widget
        self.photo_list = photos.PhotoList(self)
        self.overview_tab_layout.addWidget(self.photo_list)

        for table in (self.split_table, self.info_table, self.curve_table):
            table.set_units(self.unit_system)
        self.split_table.cellEntered.connect(self.show_split)
        self.split_table.setMouseTracking(True)
        self.curve_table.cellEntered.connect(self.show_fastest)
        self.curve_table.setMouseTracking(True)

        # Set up charts
        self.charts = charts.LineChartSet(self.unit_system, self.graphs_layout)
        self.charts.add("ele", area=True)
        self.charts.add("speed")
        self.charts.add("heartrate")
        self.charts.add("cadence")
        self.charts.add("power")
        for chart in self.charts.charts.values():
            chart.widget.mouse_moved.connect(self.show_marker)
        self.zones_chart = charts.Histogram([0], self.zones_graph, self.unit_system)
        self.curve_chart = charts.LineChart(
            self.curve_graph,
            self.unit_system,
            area=True,
            vertical_ticks=12,
            horizontal_log=True,
        )

    def update_splits(self, data):
        """Update the activity splits page."""
        self.split_table.update_data(data)

    def switch_to_summary(self):
        """Update labels, map and data box."""
        self.activity_name_label.setText(self.activity.name)
        self.flags_label.setText(" | ".join(self.activity.active_flags))
        self.description_label.setText(markdown.markdown(self.activity.description))
        self.date_time_label.setText(times.nice(self.activity.start_time))
        self.activity_type_label.setText(self.activity.sport)
        self.info_table.update_data(self.activity.stats)
        if self.activity.track.has_position_data:
            self.map_widget.setVisible(True)
            self.map_widget.show_route(self.activity.track.lat_lon_list)
            self.show_map()
        else:
            self.map_widget.setVisible(False)
        self.photo_list.show_activity_photos(self.activity)

    def switch_to_data(self):
        """Update charts."""
        if self.activity.track.has_altitude_data:
            self.charts.update_show("ele", [self.activity.track.graph("ele")])
        else:
            self.charts.hide("ele")
        self.charts.update_show("speed", [self.activity.track.graph("speed")])
        if "heartrate" in self.activity.track:
            self.charts.update_show(
                "heartrate", [self.activity.track.graph("heartrate")]
            )
        else:
            self.charts.hide("heartrate")
        if "cadence" in self.activity.track:
            self.charts.update_show("cadence", [self.activity.track.graph("cadence")])
        else:
            self.charts.hide("cadence")
        if "power" in self.activity.track:
            self.charts.update_show("power", [self.activity.track.graph("power")])
        else:
            self.charts.hide("power")

    def switch_to_splits(self):
        self.update_splits(
            self.activity.track.splits(
                splitlength=self.unit_system.units["distance"].size
            )
        )

    def switch_to_zones(self):
        zones = (
            activity_types.ZONES[self.activity.sport]
            if self.activity.sport in activity_types.ZONES
            else activity_types.ZONES[None]
        )
        zones = [self.unit_system.decode(x, "speed") for x in zones]
        self.zones_chart.set_zones(zones)
        self.zones_chart.update(self.activity.track.get_zone_durations(zones))

    @property
    def good_distances(self):
        return (
            activity_types.SPECIAL_DISTANCES[self.activity.sport]
            if self.activity.sport in activity_types.SPECIAL_DISTANCES
            else activity_types.SPECIAL_DISTANCES[None]
        )

    def switch_to_curve(self):
        table, graph, self.fastest_indices = self.activity.track.get_curve(
            self.good_distances
        )
        self.curve_chart.update([graph])
        self.curve_table.update_data(list(self.good_distances.values()), table)

    def update_page(self, page):
        """Switch to a new activity tab page."""
        if page in self.updated:
            return
        (
            self.switch_to_summary,
            self.switch_to_data,
            self.switch_to_splits,
            self.switch_to_zones,
            self.switch_to_curve,
        )[page]()
        self.updated.add(page)

    def force_update_page(self, page):
        """Update a page even if it already appears up to date."""
        if page in self.updated:
            self.updated.remove(page)
        self.update_page(page)

    def show_activity(self, new_activity):
        """Display a new activity."""
        self.activity = new_activity
        self.setWindowTitle(f"Analysing {self.activity.name}")
        if self.activity.track.manual:
            self.activity_tabs.setCurrentIndex(0)
        for page in range(1, 5):
            self.activity_tabs.setTabEnabled(page, not self.activity.track.manual)
        # Previously generated pages need refreshing
        self.updated = set()
        self.update_page(self.activity_tabs.currentIndex())

    def show_map(self):
        """
        Take back the map widget.

        This is necessary because the map widget must be shared between
        all layouts, and a widget cannot be in multiple places at once.
        Call this when the activity view becomes visible.
        """
        self.map_container.addWidget(self.map_widget)

    def show_marker(self, values):
        self.map_widget.remove_highlight()
        distance = values.x()
        self.charts.show_line(distance)
        distance = self.unit_system.decode(distance, "distance")
        point = self.activity.track.lat_lng_from_distance(distance)
        self.map_widget.show_marker(point)

    def show_split(self, split, _):
        self.map_widget.remove_marker()
        self.map_widget.show_highlight(
            self.activity.track.part_lat_lon_list(
                self.unit_system.decode(split, "distance"),
                self.unit_system.decode(split + 1, "distance"),
            )
        )

    def show_fastest(self, row, _):
        self.curve_chart.set_vertical_line(
            self.unit_system.encode(list(self.good_distances)[row], "distance")
        )
        self.map_widget.remove_marker()
        section = self.fastest_indices[row]
        self.map_widget.show_highlight(
            self.activity.track.lat_lon_list[section[0] : section[1]]
        )

    def closeEvent(self, *args, **kwargs):
        self.closed.emit()
        super().closeEvent(*args, **kwargs)



import gzip
from typing import Optional
from xml.etree import ElementTree


def try_multi(point, locations) -> Optional[str]:
    """Try to search for a value in several positions."""
    for location in locations:
        try:
            return point.find(location).text
        except AttributeError:
            continue
    return None


def get_tree(filename):
    """Load the tree, getting rid of namespaces"""
    with (gzip.open if filename.suffix == ".gz" else open)(filename) as f:
        text = f.read().lstrip()
    tree = ElementTree.ElementTree(ElementTree.fromstring(text))
    for element in tree.iter():
        _, _, element.tag = element.tag.rpartition("}")
    return tree.getroot()


def load_fields(points, fields):
    """Extract the fields from a list of points activity file."""
    result = {field: [] for field in fields}
    for point in points:
        for field in fields:
            value = None
            try:
                value = fields[field](point)
            except Exception:
                value = None
            result[field].append(value)
    return {field: result[field] for field in result if set(result[field]) != {None}}



import sys
from PyQt5.QtCore import Qt
from pathlib import Path

import pkg_resources
from PyQt5 import QtWidgets
from PyQt5.QtGui import QIcon, QPixmap

import activate.app.dialogs
import activate.app.dialogs.activity
import activate.app.dialogs.settings
from activate import (
    activity,
    activity_list,
    files,
    filetypes,
    load_activity,
    serialise,
    track,
    units,
)
from activate.app import activity_view, connect, maps, paths, settings
from activate.app.ui.main import Ui_main_window

SYNC_PROGRESS_STEPS = 1000


class MainWindow(QtWidgets.QMainWindow, Ui_main_window):
    def __init__(self, activities, *args, **kwargs):
        self.activities = activities
        super().__init__(*args, **kwargs)
        self.setupUi(self)
        paths.ensure_all_present()

        self.settings = settings.load_settings()
        self.main_tabs.setTabVisible(2, bool(self.settings.servers))

        # Create a global map widget to be used everywhere. This is
        # necessary because pyqtlet doesn't support multiple L.map
        # instances.
        self.map_widget = maps.MapWidget(self)

        self.activity_summary.setup(self.unit_system, self.map_widget)
        self.social_activity_summary.setup(self.unit_system, self.map_widget)
        self.summary.setup(self.unit_system, self.map_widget, self.activities)
        self.activity_list_table.set_units(self.unit_system)
        self.social_activity_list.set_units(self.unit_system)
        self.activity_list_table.set_units(self.unit_system)

        self.update_activity_list()
        self.activity_list_table.right_clicked.connect(self.activity_list_menu)
        self.summary.records_table.gone_to.connect(self.show_activity)

        for widget, icon_name in (
            (self.action_import, "document-open"),
            (self.action_add_manual, "document-new"),
            (self.action_edit, "document-edit"),
            (self.action_analyse, "view-statistics"),
            (self.action_add_photos, "insert-image"),
            (self.action_units, "measure"),
            (self.action_servers, "network-server"),
            (self.export_menu, "document-send"),
            (self.action_quit, "application-exit"),
        ):
            widget.setIcon(QIcon.fromTheme(icon_name))

        self.main_tab_switch(0)

    def edit_unit_settings(self):
        settings_window = activate.app.dialogs.settings.SettingsDialog()
        self.settings = settings_window.exec(self.settings, "Units")

    def edit_server_settings(self):
        settings_window = activate.app.dialogs.settings.SettingsDialog()
        self.settings = settings_window.exec(self.settings, "Servers")

        self.main_tabs.setTabVisible(2, bool(self.settings.servers))

    def add_manual_activity(self):
        manual_window = activate.app.dialogs.activity.ManualActivityDialog()
        data = manual_window.exec({})
        if data:
            self.add_activity(
                activity.Activity(
                    data["Name"],
                    data["Type"],
                    track.ManualTrack(
                        data["Start Time"],
                        data["Distance"] * 1000,
                        data["Ascent"],
                        data["Duration"],
                    ),
                    "[manual]",
                    data["Flags"],
                    data["Start Time"],
                    data["Distance"] * 1000,
                    description=data["Description"],
                )
            )

    def update_activity_list(self):
        """Make the activity list show the correct activities."""
        self.activity_list_table.setRowCount(len(self.activities))
        for i, activity_ in enumerate(self.activities):
            self.activity_list_table.set_id_row(
                activity_.activity_id, activity_.list_row, i
            )
        self.activity_list_table.resizeColumnsToContents()
        self.activity_list_table.default_sort()

    def add_activity(self, new_activity, position=0):
        """Add an activity to list."""
        activity_id = new_activity.activity_id
        activity_elements = new_activity.unload(activity_list.UnloadedActivity).list_row
        self.activities.add_activity(new_activity)
        self.activity_list_table.add_id_row(activity_id, activity_elements, position)

    def update_activity(self, selected):
        """Show a new activity on the right on the Activities page."""
        self.setUpdatesEnabled(False)
        self.activity = self.activities.get_activity(
            self.activity_list_table.item(selected, 0).activity_id
        )
        self.activity_summary.show_activity(self.activity)
        self.setUpdatesEnabled(True)

    def update_social_activity(self, selected):
        """Show a new activity on the right on the Social page."""
        cell = self.social_activity_list.item(selected, 0)
        if cell is None:
            return
        activity_id = cell.activity_id
        self.setUpdatesEnabled(False)
        try:
            self.social_activity = self.social_activities.get_activity(activity_id)
        except ValueError:
            server = next(
                s
                for s in self.settings.servers
                if s.name
                in self.social_activities.by_id(activity_id).server.split("\n")
            )
            self.social_activity = activity.Activity(
                **serialise.loads(
                    server.get_data(f"get_activity/{activity_id}"), gz=True
                )
            )
            self.social_activities.provide_full_activity(
                activity_id, self.social_activity
            )
        self.social_activity_summary.show_activity(self.social_activity)
        self.setUpdatesEnabled(True)

    def import_activities(self):
        """Import some user-given activities."""
        # [1] gives file type chosen ("Activity Files (...)",
        # "All Files" etc.)
        filenames = QtWidgets.QFileDialog.getOpenFileNames(
            self,
            "Import an activity",
            str(paths.HOME),
            "Activity Files (*.gpx *.fit *.tcx *.gpx.gz *.fit.gz *.tcx.gz)",
        )[0]
        if not filenames:
            return
        self.activity_list_table.setSortingEnabled(False)
        for filename in activate.app.dialogs.progress(
            self, filenames, "Importing Activities"
        ):
            filename = Path(filename)
            try:
                self.add_activity(load_activity.import_and_load(filename, paths.TRACKS))
            except Exception as e:
                alert_box = QtWidgets.QMessageBox()
                alert_box.setText(f"Could not load {filename}:\n{e}")
                alert_box.exec()

        self.activity_list_table.setCurrentCell(0, 0)
        self.activity_list_table.setSortingEnabled(True)
        self.main_tab_switch(self.main_tabs.currentIndex())

    def export_activity(self):
        """Export the original version of the activity."""
        if files.has_extension(self.activity.original_name, ".gpx"):
            file_type = "GPX file (*.gpx)"
        elif files.has_extension(self.activity.original_name, ".fit"):
            file_type = "FIT file (*.fit)"
        elif files.has_extension(self.activity.original_name, ".tcx"):
            file_type = "TCX file (*.tcx)"
        else:
            file_type = ""
        out_name = files.decode_name(self.activity.original_name)
        filename = QtWidgets.QFileDialog.getSaveFileName(
            self, "Export Original Activity", str(paths.HOME / out_name), file_type
        )[0]

        if not filename:
            return
        self.activity.export_original(filename)

    def export_as_route(self):
        out_name = f"{self.activity.name}.gpx"
        filename = QtWidgets.QFileDialog.getSaveFileName(
            self, "Export as Route", str(paths.HOME / out_name), "GPX file (*.gpx)"
        )[0]
        if not filename:
            return
        with open(filename, "w") as f:
            f.write(filetypes.gpx.to_route(self.activity))

    def edit_activity_data(self):
        """
        Open the Edit Activity dialog.

        This then edits or deletes the activity as appropriate.
        """
        previous_sport = self.activity.sport
        edit_activity_dialog = (
            activate.app.dialogs.activity.EditManualActivityDialog()
            if self.activity.track.manual
            else activate.app.dialogs.activity.EditActivityDialog()
        )
        return_value = edit_activity_dialog.exec(self.activity)
        if not return_value:
            return
        # Delete activity
        if return_value == activate.app.dialogs.activity.DELETE_ACTIVITY:
            # Must be saved to another variable because self.activity
            # changes when the row is removed
            to_delete = self.activity
            for row in range(len(self.activities)):
                item = self.activity_list_table.item(row, 0)
                if self.activities.get_activity(item.activity_id) is to_delete:
                    self.activity_list_table.removeRow(row)
                    break
            self.activities.remove(to_delete.activity_id)
            return

        self.activity_list_table.setSortingEnabled(False)
        for row in range(len(self.activities)):
            if (
                self.activities.get_activity(
                    self.activity_list_table.item(row, 0).activity_id
                )
                == self.activity
            ):
                self.activity_list_table.set_id_row(
                    self.activity.activity_id,
                    self.activity.unload(activity_list.UnloadedActivity).list_row,
                    row,
                )
                break
        self.activities.update(self.activity.activity_id)
        self.update_activity(row)
        self.activity_list_table.setSortingEnabled(True)
        if self.activity.sport != previous_sport:
            self.summary.update_activity_types_list()

        for server in self.settings.servers:
            try:
                server.upload_activity(self.activity)
            except connect.requests.RequestException:
                continue

    def add_photos(self):
        """Open the Add Photos dialog."""
        filenames = QtWidgets.QFileDialog.getOpenFileNames(
            self,
            "Add photos",
            str(paths.HOME),
            "Image files (*.png *.jpg *.jpeg *.gif *.bmp *.ppm *.pgm *.xpm)",
        )[0]
        if not filenames:
            return

        for filename in filenames:
            self.activity.photos.append(
                files.copy_to_location_renamed(Path(filename), paths.PHOTOS)
            )
        self.activity.save(paths.ACTIVITIES)
        self.activity_summary.update()

    def show_activity(self, activity_id):
        self.main_tabs.setCurrentIndex(1)
        self.activity_list_table.select(activity_id)

    def analyse_activity(self):
        """Open a seperate window for activity analysis."""
        self.activity_view = activity_view.ActivityView()
        self.activity_view.setup(self.unit_system, self.map_widget)
        try:
            activity = (
                self.activity
                if self.main_tabs.tabText(self.main_tabs.currentIndex()) == "Activities"
                else self.social_activity
            )
        except AttributeError:
            return
        self.activity_view.show_activity(activity)
        self.activity_view.closed.connect(self.activity_view_closed)
        self.activity_view.create()
        self.activity_view.showMaximized()

    def match_activity(self):
        self.activity_list_table.setUpdatesEnabled(False)
        if self.action_match.text() == "Clear Match":
            self.action_match.setText("Find Matching")
            self.activity_list_table.filter({a.activity_id for a in self.activities})
        else:
            self.activity_list_table.filter(
                self.activities.get_matching(
                    self.activity,
                    progress=lambda x: activate.app.dialogs.progress(
                        self, x, "Finding matching activities"
                    ),
                )
            )
            self.action_match.setText("Clear Match")
        self.activity_list_table.setUpdatesEnabled(True)

    def activity_view_closed(self):
        self.activity_summary.show_map()
        self.map_widget.remove_marker()
        self.map_widget.remove_highlight()

    def main_tab_switch(self, tab):
        """
        Switch between the main tabs at the top.

        Triggers the opened tab to update.
        """
        tab_name = self.main_tabs.tabText(tab)
        for action in (self.action_edit, self.action_add_photos, self.export_menu):
            action.setEnabled(tab_name == "Activities")
        self.activity_menu.setEnabled(tab_name != "Summary")
        if tab_name == "Summary":
            self.summary.summary_tab_switch()
        elif tab_name == "Activities":
            self.activity_summary.show_map()
            if not self.activity_list_table.selectedItems():
                self.activity_list_table.selectRow(0)
        elif tab_name == "Social":
            self.social_activity_summary.show_map()
            self.social_tab_update()
        else:
            raise ValueError("Invalid tab")

    def get_social_activities(self):
        """
        Download all activities from each server.

        Gets the activity list from the server, and then downloads each
        activity. Also uploads missing activities.
        """
        self.social_activities = activity_list.ActivityList([], None)
        dialog = QtWidgets.QProgressDialog(
            "Syncing",
            "Cancel",
            0,
            len(self.settings.servers) * SYNC_PROGRESS_STEPS,
            self,
        )
        dialog.setWindowModality(Qt.WindowModal)
        dialog.forceShow()
        for i, server in enumerate(self.settings.servers):
            dialog.setLabelText(f"Getting activity list from {server.name}")
            dialog.setValue(SYNC_PROGRESS_STEPS * i)
            try:
                server_activities = activity_list.from_serial(
                    serialise.loads(server.get_data("get_activities")), None
                )
            except connect.requests.RequestException:
                continue
            own_ids = set(a.activity_id for a in self.activities)
            dialog.setValue(round(SYNC_PROGRESS_STEPS * (i + 1 / 3)))
            dialog.setLabelText(f"Syncing activities with {server.name}")
            for j, activity_ in enumerate(server_activities):
                activity_.server = server.name
                if activity_.username == server.username:
                    aid = activity_.activity_id
                    if aid not in own_ids:
                        server.get_data(f"delete_activity/{aid}")
                        continue
                    own_ids.remove(aid)
                try:
                    previous = self.social_activities.by_id(activity_.activity_id)
                    previous.server += f"\n{activity_.server}"
                    previous.username += f"\n{activity_.username}"
                except KeyError:
                    self.social_activities.append(activity_)
                dialog.setValue(
                    round(
                        SYNC_PROGRESS_STEPS
                        * (i + (1 + (j + 1) / len(server_activities)) / 3)
                    )
                )
            dialog.setValue(round(SYNC_PROGRESS_STEPS * (i + 2 / 3)))
            if not own_ids:
                continue
            dialog.setLabelText(f"Uploading activities to {server.name}")
            for j, missing_id in enumerate(own_ids):
                try:
                    server.upload_activity(self.activities.get_activity(missing_id))
                except connect.requests.RequestException:
                    break
                dialog.setValue(
                    round(SYNC_PROGRESS_STEPS * (i + (2 + (1 + j) / len(own_ids)) / 3))
                )
        dialog.setValue(SYNC_PROGRESS_STEPS * len(self.settings.servers))

    def social_tab_update(self):
        self.get_social_activities()
        self.social_tree.set_servers(self.settings.servers, self.social_activities)
        self.social_activity_list.setUpdatesEnabled(False)
        self.social_activity_list.setSortingEnabled(False)
        self.social_activity_list.setRowCount(0)
        self.social_activity_list.setRowCount(len(self.social_activities))
        for row, activity_ in enumerate(self.social_activities):
            self.social_activity_list.set_id_row(
                activity_.activity_id, activity_.list_row, row
            )
        self.filter_social_activities()
        self.social_activity_list.resizeColumnsToContents()
        self.social_activity_list.default_sort()

    def filter_social_activities(self):
        self.social_activity_list.setUpdatesEnabled(False)
        self.social_activity_list.filter_by_server(
            self.social_tree.get_enabled_servers()
        )
        self.social_activity_list.setUpdatesEnabled(True)

    @property
    def unit_system(self):
        system = units.UNIT_SYSTEMS[self.settings.unit_system]
        for dimension, unit in self.settings.custom_units.items():
            unit = units.UNIT_NAMES[unit]
            system.units[dimension] = unit
        return system

    def activity_list_menu(self, event):
        self.activity_menu.exec(event.globalPos())

    def closeEvent(self, event):
        self.activities.save()
        return super().closeEvent(event)


def main():
    """Run the app and display the main window."""
    app = QtWidgets.QApplication(sys.argv)
    icon = QPixmap()
    icon.loadFromData(
        pkg_resources.resource_string("activate.resources", "icons/icon.png")
    )
    app.setWindowIcon(QIcon(icon))
    app.setApplicationName("Activate")
    main_window = MainWindow(activity_list.from_disk(paths.DATA))

    main_window.showMaximized()
    sys.exit(app.exec())



"""Functions for manipulating files."""
import shutil
from glob import glob
from pathlib import Path


def has_extension(filename: Path, extension: str) -> bool:
    """Determine if a file path has a given extension."""
    return filename.suffix.casefold() == "." + extension


def encode_name(filename: str, current_filenames, directory: Path) -> Path:
    """
    Rename a file to avoid name collisions.

    Renames foo.gpx to _0_foo.gpx if foo.gpx already exists. If that
    exists tries _1_foo.gpx, _2_foo.gpx... until a free name is found.
    If the name already starts with an underscore, the bare name cannot
    be used, _foo_bar.gpx becomes _0__foo_bar.gpx, _1__foo_bar.gpx etc.
    """
    # No-underscore, unique filename
    if filename[0] != "_":
        full_name = directory / filename
        if full_name not in current_filenames:
            return full_name

    # Others
    i = 0
    while True:
        full_name = directory / f"_{i}_{filename}"
        i += 1
        if full_name not in current_filenames:
            return full_name


def decode_name(filename: str) -> str:
    """Get a file's original name from its encoded one."""
    if filename[0] != "_":
        return filename
    filename = filename[filename[1:].index("_") + 2 :]
    return filename


def copy_to_location_renamed(filename: Path, copy_to: Path) -> Path:
    """
    Copy a file to a location, renaming it if necessary.

    Returns the new filename.
    """
    filenames = glob(f"{copy_to}*")
    out_name = encode_name(filename.name, filenames, copy_to)
    shutil.copy2(filename, out_name)
    return out_name



import datetime

import PyQt5
from PyQt5 import QtWidgets

from activate import activity_types
from activate.app import paths
from activate.app.dialogs import FormDialog
from activate.app.widgets import ActivityFlagEdit, DurationEdit, Form

DELETE_ACTIVITY = 222  # 0xDE[lete]


class ManualActivityDialog(FormDialog):
    def __init__(self, *args, **kwargs):
        layout = {
            "Name": QtWidgets.QLineEdit(),
            "Type": QtWidgets.QComboBox(),
            "Distance": QtWidgets.QDoubleSpinBox(),
            "Start Time": QtWidgets.QDateTimeEdit(),
            "Duration": DurationEdit(),
            "Ascent": QtWidgets.QDoubleSpinBox(),
            "Flags": ActivityFlagEdit(),
            "Description": QtWidgets.QPlainTextEdit(),
        }
        layout["Type"].currentTextChanged.connect(layout["Flags"].change_options)
        layout["Type"].addItems(activity_types.TYPES)
        layout["Distance"].setRange(0, 100000)
        layout["Ascent"].setRange(0, 100000)
        super().__init__(*args, form=Form(layout), **kwargs)
        self.setWindowTitle("Manual Activity")

    def accept(self):
        if self.form.fields["Duration"].value() > datetime.timedelta(0):
            super().accept()


class EditActivityDialog(FormDialog):
    def __init__(self, *args, **kwargs):
        layout = {
            "Name": QtWidgets.QLineEdit(),
            "Type": QtWidgets.QComboBox(),
            "Flags": ActivityFlagEdit(),
            "Description": QtWidgets.QPlainTextEdit(),
        }
        layout["Type"].currentTextChanged.connect(layout["Flags"].change_options)
        layout["Type"].addItems(activity_types.TYPES)
        super().__init__(*args, form=Form(layout), **kwargs)
        self.setWindowTitle("Edit Activity")
        self.add_delete_button()

    def add_delete_button(self):
        self.delete_button = QtWidgets.QPushButton("Delete Activity")
        self.delete_button.setIcon(PyQt5.QtGui.QIcon.fromTheme("edit-delete"))
        self.delete_button.clicked.connect(self.handle_delete_button)
        self.main_layout.insertWidget(1, self.delete_button)

    def apply_to_activity(self, data):
        """Apply the settings to an self.activity."""
        self.activity.name = data["Name"]
        self.activity.sport = data["Type"]
        self.activity.description = data["Description"]
        self.activity.flags = data["Flags"]
        self.activity.save(paths.ACTIVITIES)

    def handle_delete_button(self):
        confirm_box = QtWidgets.QMessageBox()
        confirm_box.setText(f"Are you sure you want to delete {self.activity.name}?")
        confirm_box.setStandardButtons(
            QtWidgets.QMessageBox.Yes | QtWidgets.QMessageBox.No
        )
        result = confirm_box.exec()
        if result == QtWidgets.QMessageBox.Yes:
            self.done(DELETE_ACTIVITY)

    def exec(self, activity):
        self.activity = activity
        result = super().exec(
            {
                "Name": self.activity.name,
                "Type": self.activity.sport,
                "Flags": self.activity.flags,
                "Description": self.activity.description,
            }
        )
        if result and result != DELETE_ACTIVITY:
            self.apply_to_activity(result)
        return result


class EditManualActivityDialog(EditActivityDialog, ManualActivityDialog):
    def __init__(self, *args, **kwargs):
        ManualActivityDialog.__init__(self, *args, **kwargs)
        self.setWindowTitle("Edit Activity")
        self.add_delete_button()

    def apply_to_activity(self, data):
        super().apply_to_activity(data)
        self.activity.track.length = data["Distance"]
        self.activity.track.start_time = data["Start Time"]
        self.activity.track.elapsed_time = data["Duration"]
        self.activity.track.ascent = data["Ascent"]

    def exec(self, activity):
        self.activity = activity

        result = ManualActivityDialog.exec(
            self,
            {
                "Name": self.activity.name,
                "Type": self.activity.sport,
                "Distance": self.activity.track.length / 1000,
                "Start Time": self.activity.track.start_time,
                "Duration": self.activity.track.elapsed_time,
                "Ascent": self.activity.track.ascent,
                "Flags": self.activity.flags,
                "Description": self.activity.description,
            },
        )
        if result and result != DELETE_ACTIVITY:
            self.apply_to_activity(result)
        return result



"""Convert between coordinate systems."""
from math import cos, radians, sin, sqrt

EARTH_RADIUS = 6378137
E_2 = 0.00669437999014


def to_cartesian(lat, lon, ele):
    """Convert from geodetic to cartesian coordinates based on WGS 84."""
    if None in {lat, lon, ele}:
        return (None, None, None)
    lat = radians(lat)
    lon = radians(lon)
    sin_lat = sin(lat)
    cos_lat = cos(lat)
    sin_lon = sin(lon)
    cos_lon = cos(lon)
    partial_radius = EARTH_RADIUS / sqrt(1 - E_2 * sin_lat ** 2)
    lat_radius = (ele + partial_radius) * cos_lat
    return (
        lat_radius * cos_lon,
        lat_radius * sin_lon,
        ((1 - E_2) * partial_radius + ele) * sin_lat,
    )



"""Contains the Track class and functions for handling tracks."""
import math
from dataclasses import dataclass
from datetime import datetime, timedelta
from functools import lru_cache

from dtw import dtw

from activate import geometry, times
from activate.units import DimensionValue

SPEED_RANGE = 1

FIELD_DIMENSIONS = {
    "lat": "latlon",
    "lon": "latlon",
    "ele": "altitude",
    "height_change": "altitude",
    "vertical_speed": "vertical_speed",
    "climb": "altitude",
    "desc": "altitude",
    "gradient": None,
    "angle": "angle",
    "time": "time",
    "speed": "speed",
    "dist": "distance",
    "dist_to_last": "distance",
    "cadence": "cadence",
    "heartrate": "heartrate",
    "power": "power",
}

NON_SAVED = {"vertical_speed", "climb", "desc", "gradient", "angle"}


def lerp(value1, value2, ratio):
    """
    Interpolate between value1 and value2.

    lerp(x, y, 0) = x
    lerp(x, y, 0.5) = (x + y) / 2
    lerp(x, y, 1) = y
    """
    return value1 * (1 - ratio) + value2 * ratio


def infer_nones(data):
    """Infer None values by linear interpolation."""
    none_count = 0
    last_good = None
    for index, value in enumerate(data):
        if value is None:
            none_count += 1
            continue
        if none_count:
            # Write back interpolated values
            gap_size = none_count + 1
            for write_back in range(1, gap_size):
                # Nones at start
                if last_good is None:
                    data[index - write_back] = value
                # Nones in middle
                else:
                    data[index - write_back] = lerp(
                        value, last_good, write_back / gap_size
                    )

            none_count = 0
        last_good = value
    if not none_count:
        return data
    # Nones at end
    if last_good is None:
        raise ValueError("Cannot interpolate from all Nones")
    for write_back in range(none_count + 1):
        data[index - write_back] = last_good
    return data


def get_nearby_indices(length, position, number=1) -> range:
    """
    Return numbers around position, with at most number either side.

    If position is too close to 0 or length, the excess points are
    removed.
    """
    return range(max(position - number, 0), min(position + number + 1, length))


try:
    import dist
except ImportError:

    def dist(point1, point2):
        return math.sqrt(sum((c1 - c2) ** 2 for c1, c2 in zip(point1, point2)))


def enumerate_from(list_, point):
    return enumerate(list_[point:], point)


@dataclass
class ManualTrack:
    """A manual track with a few basic values."""

    start_time: datetime
    length: float
    ascent: float
    elapsed_time: timedelta

    has_altitude_data = False
    has_position_data = False
    manual = True

    def average(self, field):
        if field == "speed":
            return self.length / self.elapsed_time.total_seconds()
        raise AttributeError(f"{self.__class__.__qualname__} has no average {field}")

    def __contains__(*_):
        return False


class Track:
    """
    A series of GPS points at given times.

    A track is considered to be purely a sequence of GPS points, with
    extra data for each point. For more metadata such as a name or
    description, the Track should be wrapped in an Activity.

    Some tracks (those representing pool swims) have no position data.
    """

    manual = False

    def __init__(self, fields):
        self.fields = fields
        for essential in ("lat", "lon"):
            if essential in self.fields and set(self[essential]) != {None}:
                infer_nones(self[essential])

        if "time" in self.fields:
            infer_nones(self["time"])

        self["ele"] = (
            infer_nones(self["ele"])
            if self.has_altitude_data
            else [0 for _ in range(len(self))]
        )

    def __getitem__(self, field):
        try:
            return self.fields[field]
        except KeyError:
            if field == "dist_to_last":
                self.calculate_dist_to_last()
            elif field == "dist":
                self.calculate_dist()
            elif field == "speed":
                self.calculate_speed()
            elif field in {"climb", "desc"}:
                self.calculate_climb_desc()
            elif field == "height_change":
                self.calculate_height_change()
            elif field == "vertical_speed":
                self.calculate_vertical_speed()
            elif field == "gradient":
                self.calculate_gradient()
            elif field == "angle":
                self.calculate_angle()
        return self.fields[field]

    def __setitem__(self, field, value):
        self.fields[field] = value

    def __contains__(self, field):
        if field in {"dist_to_last", "dist"}:
            return "dist" in self or "dist_to_last" in self or self.has_position_data
        if field in {
            "climb",
            "desc",
            "height_change",
            "vertical_speed",
            "gradient",
            "angle",
        }:
            return self.has_altitude_data
        return field in self.fields

    @property
    @lru_cache(128)
    def xyz(self):
        return [
            geometry.to_cartesian(*point)
            for point in zip(self["lat"], self["lon"], self["ele"])
        ]

    def calculate_dist_to_last(self):
        """Calculate distances between adjacent points"""
        self.fields["dist_to_last"] = [None]
        if "dist" in self.fields:
            for point in range(1, len(self)):
                relevant = self.fields["dist"][point - 1 : point + 1]
                self.fields["dist_to_last"].append(
                    None if None in relevant else relevant[1] - relevant[0]
                )
        else:
            self.fields["dist_to_last"] += [
                dist(self.xyz[point], self.xyz[point - 1])
                for point in range(1, len(self))
            ]

    def calculate_climb_desc(self):
        self.fields["climb"] = [None]
        self.fields["desc"] = [None]
        for height_change in self["height_change"][1:]:
            # Not using max in order to have integers instead of floats,
            # since max(0, 0.0) (0.0 is common in tracks) is 0.0.
            self.fields["climb"].append(0 if height_change <= 0 else height_change)
            self.fields["desc"].append(0 if height_change >= 0 else -height_change)

    def calculate_dist(self):
        """Calculate cumulative distances"""
        total_dist = 0
        new_dist = [0]
        for dist in self["dist_to_last"][1:]:
            total_dist += dist
            new_dist.append(total_dist)
        self.fields["dist"] = new_dist

    def calculate_speed(self):
        """Calculate speeds at each point"""
        speeds = []
        for point_index in range(len(self)):
            relevant_points = get_nearby_indices(
                len(self), point_index, number=SPEED_RANGE
            )
            time_diff = (
                self["time"][relevant_points[-1]] - self["time"][relevant_points[0]]
            ).total_seconds()
            distance = sum(
                self["dist_to_last"][p]
                for p in relevant_points[1:]
                if self["dist_to_last"][p] is not None
            )
            if time_diff:
                speeds.append(distance / time_diff)
            elif speeds:
                speeds.append(speeds[-1])
            else:
                speeds.append(0)
        self["speed"] = speeds

    def calculate_height_change(self):
        """Calculate differences in elevation between adjacent points."""
        self.fields["height_change"] = [None]
        last = self["ele"][0]
        for current in self["ele"][1:]:
            self.fields["height_change"].append(current - last)
            last = current

    def calculate_vertical_speed(self):
        """Calculate vertical speed at each point."""
        self.fields["vertical_speed"] = [None]
        for point in range(1, len(self)):
            height_change = self["height_change"][point]
            time = self["time"][point]
            last_time = self["time"][point - 1]
            if None in {height_change, time, last_time}:
                self.fields["vertical_speed"].append(None)
            else:
                self.fields["vertical_speed"].append(
                    height_change / (time - last_time).total_seconds()
                )

    def calculate_gradient(self):
        """Calculate the gradient at each point."""
        self.fields["gradient"] = [None]
        for dist, height_change in list(
            zip(self["dist_to_last"], self["height_change"])
        )[1:]:
            self.fields["gradient"].append(
                None if dist in {None, 0} else height_change / dist
            )

    def calculate_angle(self):
        """Calculate the angle of inclination at each point."""
        self.fields["angle"] = [
            None if g is None else math.atan(g) for g in self["gradient"]
        ]

    def __len__(self):
        return len(next(iter(self.fields.values())))

    def without_nones(self, field):
        return (v for v in self[field] if v is not None)

    @lru_cache(128)
    def average(self, field):
        """Get the mean value of a field, ignoring missing values."""
        if field == "speed":
            return self.length / self.elapsed_time.total_seconds()

        valid = list(self.without_nones(field))
        return sum(valid) / len(valid)

    @lru_cache(128)
    def maximum(self, field):
        """Get the maximum value of a field, ignoring missing values."""
        return max(self.without_nones(field))

    # Caching necessary to avoid fake elevation data
    @property
    @lru_cache(128)
    def has_altitude_data(self):
        return "ele" in self.fields

    @property
    @lru_cache(128)
    def has_position_data(self):
        return "lat" in self.fields and "lon" in self.fields

    @property
    @lru_cache(128)
    def lat_lon_list(self):
        return [[x, y] for x, y in zip(self["lat"], self["lon"])]

    def part_lat_lon_list(self, min_dist, max_dist):
        track = []
        for dist, lat, lon in zip(self["dist"], self["lat"], self["lon"]):
            if dist is None or dist < min_dist:
                continue
            if dist >= max_dist:
                return track
            track.append([lat, lon])
        return track

    @property
    @lru_cache(128)
    def ascent(self):
        if self.has_altitude_data:
            return sum(self.without_nones("climb"))

    @property
    @lru_cache(128)
    def descent(self):
        if self.has_altitude_data:
            return sum(self.without_nones("desc"))

    @property
    def start_time(self):
        return self["time"][0]

    @property
    @lru_cache(128)
    def elapsed_time(self) -> timedelta:
        end_time = self["time"][-1]
        return end_time - self.start_time

    @property
    @lru_cache(128)
    def moving_time(self) -> timedelta:
        total_time = timedelta(0)
        last_distance = 0
        last_time = self.start_time
        for distance, time in zip(self["dist"][1:], self["time"][1:]):
            if distance is None:
                continue
            time_difference = time - last_time
            if not time_difference:
                continue
            distance_difference = distance - last_distance
            if distance_difference < 1:
                continue
            if distance_difference / time_difference.total_seconds() > 0.2:
                total_time += time_difference
            elif distance < last_distance:
                raise ValueError("Distance increase")
            last_distance = distance
            last_time = time

        return total_time

    @property
    @lru_cache(128)
    def average_speed_moving(self):
        return self.length / self.moving_time.total_seconds()

    @property
    @lru_cache(128)
    def distance_in_days(self) -> dict:
        if self.start_time.date() == self["time"][-1].date():
            return {self.start_time.date(): self.length}
        last_time = self.start_time
        last_date = last_time.date()
        totals = {last_date: 0}

        for dist_to_last, time in zip(self["dist_to_last"][1:], self["time"][1:]):
            if dist_to_last is None:
                continue
            date = time.date()
            if date == last_date:
                totals[last_date] += dist_to_last
            else:
                speed = dist_to_last / (time - last_time).total_seconds()
                totals[last_date] += (
                    speed * (times.end_of(last_time, "day") - last_time).total_seconds()
                )
                for days in range(1, (date - last_date).days):
                    day = last_date + timedelta(days)
                    totals[day] = speed * timedelta(days=1)
                totals[date] = (
                    speed * (time - times.start_of(time, "day")).total_seconds()
                )
                last_date = date

            last_time = time
        return totals

    def lat_lng_from_distance(self, distance):
        point0 = 0
        for point1, dist1 in enumerate(self["dist"]):
            if dist1 is None:
                continue
            if dist1 > distance:
                break
            point0 = point1
        else:
            return None

        dist0 = self["dist"][point0]
        lat0 = self["lat"][point0]
        lon0 = self["lon"][point0]
        if dist0 == dist1:
            return (lat0, lon0)
        lat1 = self["lat"][point1]
        lon1 = self["lon"][point1]
        ratio = (distance - dist0) / (dist1 - dist0)
        return (lerp(lat0, lat1, ratio), lerp(lon0, lon1, ratio))

    def graph(self, y_data, x_data="dist") -> tuple:
        """Get x and y data as (data, dimension) tuples."""
        return (
            (self[x_data], FIELD_DIMENSIONS[x_data]),
            (self[y_data], FIELD_DIMENSIONS[y_data]),
        )

    @property
    def length(self):
        return next(x for x in reversed(self["dist"]) if x is not None)

    def splits(self, splitlength=1000) -> list:
        """
        Split an activity into splits, with per-split data.

        Each split is a list in the format
        [lap, split, speed, net climb, total climb].
        """
        splits = []
        lasttime = None
        lastalt = None
        total_climb = 0
        for time, dist, alt, climb in zip(
            self["time"], self["dist"], self["ele"], self["climb"]
        ):
            if lasttime is None:
                lasttime = time
                lastalt = alt
            if dist is None:
                continue
            if dist // splitlength > len(splits):
                speed = splitlength / (time - lasttime).total_seconds()
                splits.append(
                    [
                        DimensionValue(time - lasttime, "time"),
                        DimensionValue(time - self.start_time, "time"),
                        DimensionValue(speed, "speed"),
                        DimensionValue(alt - lastalt, "altitude"),
                        DimensionValue(total_climb, "altitude"),
                    ]
                )
                total_climb = 0
                lasttime = None
            if climb is not None:
                total_climb += climb

        return splits

    def get_zone_durations(self, zones, field="speed", count_field="time"):
        """
        Get durations for the zones graph.

        Between all zones values, calculate the total amount of
        count_field at each point where field is within the range given
        by that pair of zones values. The zones must be sorted in
        ascending order.
        """
        buckets = {z: 0 for z in zones}
        for point in range(len(self)):
            # Work out the amount of count_field at the point
            nearby_points = get_nearby_indices(len(self), point)
            duration = (
                self[count_field][nearby_points[-1]]
                - self[count_field][nearby_points[0]]
            ) / 2
            duration = times.to_number(duration)

            # Add it to the correct bucket
            value = self[field][point]
            if value is None:
                continue
            for zone in zones[::-1]:
                if value > zone:
                    buckets[zone] += duration
                    break

        return buckets

    def get_curve(self, table_distances):
        """Get the curve and curve table for an activity."""
        table_distances = [x for x in table_distances if x <= self.length]
        time_values = [t.timestamp() for t in self["time"]]

        bests = []
        point_indices = []
        for distance in table_distances:
            for last_point, last_dist in enumerate(self["dist"]):
                if last_dist is not None and last_dist > distance:
                    break
            best = time_values[last_point] - self.start_time.timestamp()
            first_point = 0
            point = (first_point, last_point)
            for last_point, last_dist in enumerate_from(self["dist"], last_point + 1):
                if last_dist is None:
                    continue
                last_good_first_point = first_point
                for first_point, first_dist in enumerate_from(
                    self["dist"], first_point
                ):
                    if first_dist is None:
                        continue
                    if last_dist - first_dist >= distance:
                        last_good_first_point = first_point
                    else:
                        break
                first_point = last_good_first_point
                time_taken = time_values[last_point] - time_values[first_point]
                if time_taken < best:
                    best = time_taken
                    point = (first_point, last_point)
            bests.append(best)
            point_indices.append(point)

        speeds = [
            distance / time if time else 0
            for distance, time in zip(table_distances, bests)
        ]
        bests_table = [
            (
                DimensionValue(distance, "distance"),
                DimensionValue(timedelta(seconds=time), "time"),
                DimensionValue(speed, "speed"),
            )
            for distance, time, speed in zip(table_distances, bests, speeds)
        ]
        return (
            bests_table,
            ((table_distances, "distance"), (speeds, "speed")),
            point_indices,
        )

    def match(self, other):
        return dtw(self.xyz, other.xyz, distance_only=True).normalizedDistance < 40

    @property
    def save_data(self):
        return {
            key: value for key, value in self.fields.items() if key not in NON_SAVED
        }



from dataclasses import dataclass

import requests

from activate import serialise


@dataclass
class Server:
    address: str
    name: str
    username: str
    password: str

    def post_data(self, page, data) -> bytes:
        """Send a POST request to address/page with data."""
        r = requests.post(self.api_address(page), data=data, auth=self.auth)
        r.raise_for_status()
        return r.content

    def get_data(self, page) -> bytes:
        """Send a GET request to address/page."""
        r = requests.get(self.api_address(page), auth=self.auth)
        r.raise_for_status()
        return r.content

    def upload_activity(self, activity):
        self.post_data(
            "send_activity", {"activity": serialise.dump_bytes(activity.save_data)}
        )

    def api_address(self, page):
        return f"{self.address.rstrip('/')}/api/{page}"

    @property
    def auth(self):
        return requests.auth.HTTPBasicAuth(self.username, self.password)



"""Functions for loading activities from files."""
from pathlib import Path

from activate import activity, files, filetypes, track

ACTIVITY_TYPE_NAMES = {
    "run": "Run",
    "running": "Run",
    "ride": "Ride",
    "cycling": "Ride",
    "biking": "Ride",
    "walk": "Walk",
    "walking": "Walk",
    "hiking": "Walk",
    "ski": "Ski",
    "alpine_skiing": "Ski",
    "swim": "Swim",
    "swimming": "Swim",
    "row": "Row",
    "rowing": "Row",
    "1": "Ride",
    "9": "Run",
    "10": "Walk",
    "16": "Swim",
    # 28 : "Rock Climbing",
}


def convert_activity_type(activity_type: str, name) -> str:
    """Get the correct activity type from a raw one or by inference."""
    activity_type = activity_type.casefold()
    if activity_type in ACTIVITY_TYPE_NAMES:
        return ACTIVITY_TYPE_NAMES[activity_type]
    if activity_type in {"unknown", "generic"}:
        # Infer activity type from name
        for activity_type_name in ACTIVITY_TYPE_NAMES:
            if activity_type_name.isnumeric():
                continue
            if activity_type_name in name.casefold():
                return ACTIVITY_TYPE_NAMES[activity_type_name]
    return "Other"


def default_name(filename: Path):
    """Generate a default activity name from a file name."""
    return files.decode_name(filename.stem.split(".")[0])


def load(filename: Path) -> dict:
    """
    Get {"name": name, "sport": sport, "track": Track} by loading a file.

    Uses the appropriate track loader from the filetypes module.
    """
    filetype = (
        filename.with_suffix("").suffix
        if filename.suffix.casefold() == ".gz"
        else filename.suffix
    ).casefold()
    if filetype == ".gpx":
        data = filetypes.gpx.load_gpx(filename)
    elif filetype == ".fit":
        data = filetypes.fit.load_fit(filename)
    elif filetype == ".tcx":
        data = filetypes.tcx.load_tcx(filename)
    else:
        raise ValueError("Invalid filetype")

    return {
        "name": data[0] if data[0] is not None else default_name(filename),
        "sport": convert_activity_type(data[1], data[0]),
        "track": track.Track(data[2]),
    }


def import_and_load(filename, copy_to) -> activity.Activity:
    """Import an activity and copy it into the originals directory."""
    filename = files.copy_to_location_renamed(filename, copy_to)
    return activity.from_track(**load(filename), filename=filename)



"""Contains classes and for handling units and some predefined units."""
import math
from dataclasses import dataclass
from datetime import timedelta
from functools import wraps
from typing import Union


class DimensionError(Exception):
    pass


def compatible_dimensions(function):
    @wraps(function)
    def wrapper(var1, var2, **kwargs):
        if var1.dimension != var2.dimension:
            raise DimensionError(
                f"incompatible dimensions: {var1.dimension} and {var2.dimension}"
            )
        return function(var1, var2, **kwargs)

    return wrapper


@dataclass(frozen=True)
class DimensionValue:
    """A value with a dimension attached."""

    value: Union[float, timedelta]
    dimension: str

    def format(self, unit_system):
        return unit_system.format(self.value, self.dimension)

    def encode(self, unit_system):
        return unit_system.encode(self.value, self.dimension)

    @compatible_dimensions
    def __lt__(self, other):
        return self.value < other.value

    @compatible_dimensions
    def __gt__(self, other):
        return self.value > other.value

    @compatible_dimensions
    def __eq__(self, other):
        return self.value == other.value

    @compatible_dimensions
    def __ne__(self, other):
        return self.value != other.value

    @compatible_dimensions
    def __le__(self, other):
        return self.value <= other.value

    @compatible_dimensions
    def __ge__(self, other):
        return self.value >= other.value

    @compatible_dimensions
    def __add__(self, other):
        return DimensionValue(self.value + other.value, self.dimension)

    @compatible_dimensions
    def __sub__(self, other):
        return DimensionValue(self.value - other.value, self.dimension)

    def __neg__(self):
        return DimensionValue(-self.value, self.dimension)


@dataclass
class Unit:
    name: str
    symbol: str
    size: float

    def encode(self, value):
        return value / self.size

    def decode(self, value):
        return value * self.size

    def format(self, value):
        return (self.encode(value), self.symbol)

    def __truediv__(self, other):
        return Unit(
            f"{self.name} per {other.name}",
            f"{self.symbol} ∕ {other.symbol}",
            self.size / other.size,
        )

    def __mul__(self, other):
        return Unit(
            f"{self.name} {other.name}",
            f"{self.symbol} {other.symbol}",
            self.size * other.size,
        )


@dataclass
class UnitConfig:
    """The current preferred unit system."""

    units: dict

    def encode(self, value, dimension):
        return self.units[dimension].encode(value)

    def decode(self, value, dimension):
        return self.units[dimension].decode(value)

    def format(self, value, dimension):
        return self.units[dimension].format(value)

    def format_name_unit(self, dimension, symbol=None, name=None) -> str:
        """
        Get 'Distance (m)' or similar string.

        Returns the dimension if it isn't recognised.
        """
        if symbol is None:
            if dimension not in self.units:
                return dimension
            symbol = self.units[dimension].symbol
        if symbol:
            return f"{dimension.title() if name is None else name} ({symbol})"
        return dimension.title()


class PaceUnit(Unit):
    """A unit of pace (1 / speed), such as 4:00 kilometres."""

    def __init__(self, distance: Unit):
        self.distance = distance
        self.name = f"minutes per {distance.name}"
        self.size = 1 / distance.size
        self.symbol = "∕ " + distance.symbol

    def encode(self, value) -> timedelta:
        return timedelta(seconds=super().encode(value))

    def __repr__(self):
        return f"{self.__class__.__qualname__}({self.distance!r})"


class DateUnit(Unit):
    """A unit giving a date."""

    def __init__(self):
        super().__init__("date", "", 1)

    def encode(self, value):
        return value.timestamp()

    def format(self, value):
        return str(value)


KM = Unit("kilometre", "km", 1000)
MILE = Unit("mile", "mi", 1609.344)
METRE = Unit("metre", "m", 1)
FOOT = Unit("foot", "ft", 0.3048)
YARD = Unit("yard", "yd", 0.9144)
SECOND = Unit("second", "s", 1)
MINUTE = Unit("minute", "min", 60)
HOUR = Unit("hour", "h", 3600)
METRE_PER_SECOND = METRE / SECOND
KM_PER_HOUR = KM / HOUR
MILE_PER_HOUR = MILE / HOUR
MILE_PER_HOUR.symbol = "mph"
FOOT_PER_MINUTE = FOOT / MINUTE
METRE_PER_MINUTE = METRE / MINUTE
TIME = Unit("minutes and seconds", "", 1)
MIN_PER_KM = PaceUnit(KM)
MIN_PER_MILE = PaceUnit(MILE)
BEAT_PER_MINUTE = Unit("beat per minute", "bpm", 1 / 60)
CYCLES_PER_MINUTE = Unit("cycles per minute", "rpm", 1 / 60)
HERTZ = Unit("hertz", "Hz", 1)
WATT = Unit("watt", "W", 1)
HORSE_POWER = Unit("horsepower", "hp", 745.6998715822702)
DEGREE = Unit("degree", "°", math.tau / 360)
RADIAN = Unit("radian", "", 1)
UNITLESS = Unit("", "", 1)
DATE = DateUnit()
ALL = {
    "distance": (METRE, FOOT, YARD, KM, MILE),
    "altitude": (METRE, FOOT),
    "speed": (METRE_PER_SECOND, KM_PER_HOUR, MILE_PER_HOUR, METRE_PER_MINUTE),
    "vertical_speed": (METRE_PER_SECOND, METRE_PER_MINUTE, FOOT_PER_MINUTE),
    "real_time": (SECOND, MINUTE, HOUR),
    "time": (TIME,),
    "pace": (MIN_PER_KM, MIN_PER_MILE),
    "date": (DATE,),
    "heartrate": (BEAT_PER_MINUTE,),
    "cadence": (CYCLES_PER_MINUTE,),
    "power": (WATT,),
    "angle": (DEGREE,),
    None: (UNITLESS,),
}

METRIC = {
    "distance": KM,
    "altitude": METRE,
    "speed": KM_PER_HOUR,
    "vertical_speed": METRE_PER_MINUTE,
    "real_time": MINUTE,
    "time": TIME,
    "pace": MIN_PER_KM,
    "date": DATE,
    "heartrate": BEAT_PER_MINUTE,
    "cadence": CYCLES_PER_MINUTE,
    "power": WATT,
    "angle": DEGREE,
    None: UNITLESS,
}

IMPERIAL = {
    "distance": MILE,
    "altitude": FOOT,
    "speed": MILE_PER_HOUR,
    "vertical_speed": FOOT_PER_MINUTE,
    "real_time": MINUTE,
    "time": TIME,
    "pace": MIN_PER_MILE,
    "date": DATE,
    "heartrate": BEAT_PER_MINUTE,
    "cadence": CYCLES_PER_MINUTE,
    "power": HORSE_POWER,
    "angle": DEGREE,
    None: UNITLESS,
}

DEFAULT = "Metric"

UNIT_SYSTEMS = {"Metric": UnitConfig(METRIC), "Imperial": UnitConfig(IMPERIAL)}

UNIT_NAMES = {
    u.name: u
    for u in (
        KM,
        MILE,
        METRE,
        FOOT,
        YARD,
        SECOND,
        MINUTE,
        HOUR,
        METRE_PER_SECOND,
        KM_PER_HOUR,
        MILE_PER_HOUR,
        FOOT_PER_MINUTE,
        METRE_PER_MINUTE,
        TIME,
        MIN_PER_KM,
        MIN_PER_MILE,
        BEAT_PER_MINUTE,
        CYCLES_PER_MINUTE,
        HERTZ,
        WATT,
        HORSE_POWER,
        DEGREE,
        RADIAN,
        UNITLESS,
        DATE,
    )
}



"""Custom QCharts."""
import math
from collections import namedtuple
from datetime import datetime, timedelta
from itertools import zip_longest

from PyQt5 import QtChart, QtWidgets
from PyQt5.QtCore import QPointF, Qt, pyqtSignal
from PyQt5.QtGui import QPainter

from activate import number_formats, times, units


def axis_number_format(axis):
    """Format axis labels with the correct number of decimal places."""
    interval = (axis.max() - axis.min()) / (axis.tickCount() - 1)
    if interval.is_integer():
        axis.setLabelFormat("%i")
    else:
        axis.setLabelFormat(f"%.{max(0, -math.floor(math.log10(interval)))}f")


def date_axis_format(difference: timedelta) -> str:
    """Get the formatting for a date axis based on its range."""
    if difference >= timedelta(days=365):
        return "MMM yyyy"
    if difference >= timedelta(days=100):
        return "MMMM"
    if difference >= timedelta(days=5):
        return "dd MMMM"
    if difference >= timedelta(days=3):
        return "hh:00 d MMM"
    if difference >= timedelta(days=1):
        return "hh:mm d MMM"
    if difference >= timedelta(hours=12):
        return "hh:mm"
    return "hh:mm:ss"


class MinMax:
    """Keeps track of the minimum and maximum of some data."""

    def __init__(self, *args):
        if args:
            try:
                self.minimum = min(min(a) for a in args if a)
                self.maximum = max(max(a) for a in args if a)
                return
            # No values given
            except ValueError:
                pass
        self.minimum = None
        self.maximum = None

    def update(self, value):
        """Add a new value to the MinMax."""
        if self.minimum is None:
            self.minimum = value
            self.maximum = value

        self.minimum = min(self.minimum, value)
        self.maximum = max(self.maximum, value)

    @property
    def range(self):
        if self.minimum is None:
            return None
        return self.maximum - self.minimum

    @property
    def ratio(self):
        if self.minimum is None:
            return None
        return self.maximum / self.minimum

    def __bool__(self):
        return self.minimum is not None and self.range > 0

    def __repr__(self):
        if self.minimum is None:
            return f"{self.__class__.__qualname__}()"
        return f"{self.__class__.__qualname__}(({self.minimum!r}, {self.maximum!r}))"


def data_to_points(data):
    """Convert a [series1, series2] of data to a list of QPointF."""
    return [QPointF(*p) for p in zip(*data)]


series_gc_prevent = []


def create_axis(log=False):
    if log:
        axis = QtChart.QLogValueAxis()
        axis.setMinorTickCount(-1)
        axis.setLabelFormat("%g")
        return axis
    return QtChart.QValueAxis()


class ChartView(QtChart.QChartView):
    mouse_moved = pyqtSignal(QPointF)

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.setMouseTracking(True)

    def mouseMoveEvent(self, event):
        self.mouse_moved.emit(self.chart().mapToValue(event.localPos()))
        super().mouseMoveEvent(event)


class Chart(QtChart.QChart):
    """A chart with sensible defaults and extra functionality."""

    def __init__(
        self,
        seriess,
        widget,
        unit_system,
        title=None,
        horizontal_ticks=12,
        vertical_ticks=5,
        horizontal_log=False,
        vertical_log=False,
        x_axis_label=True,
        y_axis_label=True,
    ):
        """Create a new chart."""
        self.widget = widget
        self.unit_system = unit_system
        self.horizontal_ticks = horizontal_ticks
        self.vertical_ticks = vertical_ticks
        self.horizontal_log = horizontal_log
        self.vertical_log = vertical_log
        self.x_axis_label = x_axis_label
        self.y_axis_label = y_axis_label
        super().__init__()
        self.setAnimationOptions(self.SeriesAnimations)
        self.widget.setRenderHint(QPainter.Antialiasing, True)
        self.legend().hide()

        x_axis = create_axis(horizontal_log)
        y_axis = create_axis(vertical_log)
        self.addAxis(x_axis, Qt.AlignBottom)
        self.addAxis(y_axis, Qt.AlignLeft)
        for series in seriess:
            self.addSeries(series)
            series.attachAxis(x_axis)
            series.attachAxis(y_axis)
        self.widget.setChart(self)
        if title is not None:
            self.setTitle(title)

    def set_axis_dimensions(self, x_axis_dimension, y_axis_dimension):
        if x_axis_dimension and self.x_axis_label:
            self.x_axis.setTitleText(
                self.unit_system.format_name_unit(x_axis_dimension)
            )

        if y_axis_dimension and self.y_axis_label:
            self.y_axis.setTitleText(
                self.unit_system.format_name_unit(y_axis_dimension)
            )

    def update_axis(self, direction, ticks, minimum, maximum):
        """Change an axis range to fit minimum and maximum."""
        axis = self.axes(direction)[0]
        if isinstance(axis, TimePeriodAxis):
            axis.setRange(minimum, maximum)
        elif isinstance(axis, QtChart.QValueAxis):
            fake_axis = QtChart.QValueAxis()
            fake_axis.setRange(minimum, maximum)
            fake_axis.setTickCount(ticks)
            fake_axis.applyNiceNumbers()
            axis.setRange(fake_axis.min(), fake_axis.max())
            axis.setTickCount(fake_axis.tickCount())
            axis_number_format(axis)
        elif isinstance(axis, QtChart.QLogValueAxis):
            # Minimum must be decreased slightly to add the necessary extra tick
            axis.setRange(minimum / 1.00001, maximum)
        # For date axes in subclass
        else:
            axis.setRange(minimum, maximum)
            axis.setTickCount(ticks)

    def add_legend(self, names):
        self.legend().show()
        for name, series in zip(names, self.series()):
            series.setName(name)

    def remove_legend(self):
        self.legend().hide()

    def set_vertical_line(self, x):
        x = self.mapToPosition(QPointF(x, 0)).x()
        try:
            self.line.setLine(x, 0, x, 999)
        except AttributeError:
            line = QtWidgets.QGraphicsLineItem(x, 0, x, 999)
            self.widget.scene().addItem(line)
            self.line = line

    @property
    def x_axis(self):
        return self.axes(Qt.Horizontal)[0]

    @property
    def y_axis(self):
        return self.axes(Qt.Vertical)[0]


class LineChart(Chart):
    """A chart with 1+ QLineSeries on it."""

    def __init__(self, widget, unit_system, area=False, series_count=1, **kwargs):
        """Add a line chart to widget."""
        seriess = []
        for _ in range(series_count):
            series = QtChart.QLineSeries()
            if area:
                area = QtChart.QAreaSeries()
                area.setUpperSeries(series)
                # Save series so it doesn't get garbage collected
                series_gc_prevent.append(series)
                series = area
            seriess.append(series)
        super().__init__(seriess, widget, unit_system, **kwargs)

    def encode_data(self, data):
        """Convert data with a dimension to floats with correct units."""
        # Convert units
        data = zip(
            *(
                [
                    None if x is None else self.unit_system.encode(x, unit)
                    for x in series
                ]
                for series, unit in data
            )
        )
        # Get rid of Nones
        return list(zip(*(p for p in data if None not in p)))

    @property
    def data_series(self):
        return [
            s.upperSeries() if isinstance(s, QtChart.QAreaSeries) else s
            for s in self.series()
        ]

    def clear(self):
        for series in self.data_series:
            series.setVisible(False)

    def update(self, data):
        """
        Change a line chart's data.

        Input format:
        (((x_values_0, x_unit), (y_values_0, y_unit)),
         ((x_values_1, x_unit), (y_values_1, y_unit)),
         ...
        )
        """
        if not data:
            for series in self.data_series:
                series.setVisible(False)
            return
        x_dimension = data[0][0][1]
        y_dimension = data[0][1][1]
        self.set_axis_dimensions(x_dimension, y_dimension)
        # Convert to the correct units
        data = [self.encode_data(d) for d in data]
        # Extract 'real' series from an area chart
        x_range = MinMax(*(d[0] for d in data))
        y_range = MinMax(*(d[1] for d in data))
        for data_part, series in zip_longest(data, self.data_series):
            if data_part is None:
                series.setVisible(False)
            else:
                series.setVisible(True)
                series.replace(data_to_points(data_part))

        # Snap axis minima to zero
        if (
            not self.horizontal_log
            and x_range
            and x_range.minimum != 0
            and x_range.ratio > 3
        ):
            x_range.minimum = 0
        if (
            not self.vertical_log
            and y_range
            and y_range.minimum != 0
            and y_range.ratio > 3
        ):
            y_range.minimum = 0

        self.update_axis(
            Qt.Horizontal, self.horizontal_ticks, x_range.minimum, x_range.maximum
        )
        self.update_axis(
            Qt.Vertical, self.vertical_ticks, y_range.minimum, y_range.maximum
        )


class LineChartSet:
    """A set of line charts that can be hidden and shown."""

    def __init__(self, unit_system, container):
        self.unit_system = unit_system
        self.container = container
        self.charts = {}
        self.chart_views = {}

    def add(self, name, area=False):
        self.chart_views[name] = ChartView(self.container.parentWidget())
        self.charts[name] = LineChart(
            self.chart_views[name], self.unit_system, area=area
        )
        self.container.addWidget(self.chart_views[name], 1)

    def __getitem__(self, name):
        return self.charts[name]

    def show(self, name):
        self.chart_views[name].setVisible(True)

    def hide(self, name):
        self.chart_views[name].setVisible(False)

    def update_show(self, name, data):
        self[name].update(data)
        self.show(name)

    def show_line(self, x_value):
        for chart in self.charts:
            self.charts[chart].set_vertical_line(x_value)

    def __repr__(self):
        return f"<LineChartSet charts={self.charts!r}>"


class Histogram(Chart):
    def __init__(self, zones, widget, unit_system):
        """Create a histogram."""
        series = QtChart.QBarSeries()
        bar_set = QtChart.QBarSet("", series)
        series.append(bar_set)
        series.setBarWidth(1)
        super().__init__([series], widget, unit_system)
        self.set_zones(zones)

    def set_zones(self, zones):
        # Use QBarCategoryAxis instead of QCategoryAxis because the
        # latter allows putting values between categoreies instead of
        # centring them.
        cat_axis = self.x_axis
        self.removeAxis(cat_axis)
        cat_axis = QtChart.QCategoryAxis(self)
        cat_axis.setLabelsPosition(QtChart.QCategoryAxis.AxisLabelsPositionOnValue)
        # Hide the start value because zones[0] does its job
        cat_axis.setStartValue(float("-inf"))

        # Add initial label, handling negative infinity.
        if zones[0] == float("-inf"):
            cat_axis.append("\u2212\u221e", -0.5)
        else:
            zone_num = self.unit_system.encode(zones[0], "speed")
            cat_axis.append(number_formats.maybe_as_int(zone_num), -0.5)

        # Add axis labels
        for position, zone in enumerate(zones[1:-1]):
            zone_num = self.unit_system.encode(zone, "speed")
            cat_axis.append(number_formats.maybe_as_int(zone_num), position + 0.5)

        # Add final label. This should usually be infinity.
        if zones[-1] == float("inf"):
            cat_axis.append("\u221e", len(zones) - 1.5)
        else:
            zone_num = self.unit_system.encode(zones[-1], "speed")
            cat_axis.append(number_formats.maybe_as_int(zone_num), len(zones) - 1.5)

        cat_axis.setRange(-0.5, len(zones) - 1.5)

        # One less bar than there are zones borders
        series = self.series()[0]
        bar_set = series.barSets()[0]
        if bar_set.count() > len(zones) - 1:
            bar_set.remove(0, bar_set.count() - len(zones) + 1)
        elif bar_set.count() < len(zones) - 1:
            for _ in range(len(zones) - 1 - bar_set.count()):
                bar_set.append(0)
        self.addAxis(cat_axis, Qt.AlignBottom)
        series.attachAxis(cat_axis)
        self.set_axis_dimensions("speed", "Time (min)")

    def update(self, data):
        """Update the histogram data."""
        series = self.series()[0]
        bar_set = series.barSets()[0]
        for position, amount in enumerate(data.values()):
            bar_set.replace(position, units.MINUTE.encode(amount))

        # Format the vertical axis
        self.update_axis(Qt.Vertical, 15, 0, units.MINUTE.encode(max(data.values())))


class TimePeriodAxis(QtChart.QCategoryAxis):
    def __init__(self, *args, **kwargs):
        self.mode = "day"
        super().__init__(*args, **kwargs)

    def setRange(self, minimum, maximum):
        super().setRange(minimum.timestamp(), maximum.timestamp())

    def update_labels(self, minimum, maximum):
        if self.mode == "auto":
            if maximum - minimum > times.ONE_DAY * 500:
                mode = "year"
            elif maximum - minimum > times.ONE_DAY * 50:
                mode = "month"
            elif maximum - minimum > times.ONE_DAY * 10:
                mode = "day"
            else:
                mode = "weekday"
        else:
            mode = self.mode
        new_labels = times.get_periods(minimum, maximum, mode)
        for label in self.categoriesLabels():
            self.remove(label)
        for position, label in new_labels:
            self.append(label, position.timestamp())


class TimePeriodLineChart(LineChart):
    """A line chart with datetimes on the x axis."""

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.removeAxis(self.x_axis)
        self.period_axis = TimePeriodAxis()
        self.addAxis(self.period_axis, Qt.AlignBottom)
        for series in self.series():
            series.attachAxis(self.period_axis)

    def encode_data(self, data):
        """
        Convert the data provided to a more usable format.

        Input format: ((x_values, "time"), (y_values, y_format))
        Output format: (x_values, y_values)
        The output values are in the correct units for display on the
        chart.
        """
        x_data = [self.unit_system.encode(x, data[0][1]) for x in data[0][0]]
        y_data = [self.unit_system.encode(x, data[1][1]) for x in data[1][0]]
        return (x_data, y_data)

    def update_axis(self, direction, ticks, minimum, maximum):
        """Resize the chart axes."""
        if minimum is None:
            return
        if direction == Qt.Horizontal:
            minimum = datetime.fromtimestamp(minimum)
            maximum = datetime.fromtimestamp(maximum)
            # self.period_axis.setFormat(axis_format(maximum - minimum))
            self.period_axis.update_labels(minimum, maximum)
        super().update_axis(direction, ticks, minimum, maximum)


class ProgressionChart(TimePeriodLineChart):
    """A chart to show total distance, time or climbing over time."""

    YOption = namedtuple("YOption", ("function", "unit"))
    Y_OPTIONS = {
        "Distance": YOption(lambda a: a.distance, "distance"),
        "Time": YOption(lambda a: a.duration.total_seconds(), "real_time"),
        "Climb": YOption(lambda a: a.climb, "altitude"),
    }

    def update(self, summary_period, allowed_activity_types, y_mode, now, activities):
        periods, data = activities.get_progression_data(
            allowed_activity_types, summary_period, now, self.Y_OPTIONS[y_mode].function
        )
        if summary_period == "All Time":
            self.period_axis.mode = "auto"
            self.remove_legend()
        else:
            self.period_axis.mode = {
                "Year": "month",
                "Month": "day",
                "Week": "weekday",
            }[summary_period]
            self.add_legend(periods)
        super().update(
            [((d[0], "date"), (d[1], self.Y_OPTIONS[y_mode].unit)) for d in data]
        )


class FullProgressionChart(QtWidgets.QWidget):
    y_changed = pyqtSignal()

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.main_layout = QtWidgets.QHBoxLayout(self)
        self.box = QtWidgets.QComboBox(self)
        self.box.currentTextChanged.connect(self.change_y)
        self.main_layout.addWidget(self.box)
        self.chart_view = QtChart.QChartView(self)
        self.main_layout.addWidget(self.chart_view)

    def set_units(self, unit_system):
        self.unit_system = unit_system
        self.chart = ProgressionChart(
            self.chart_view,
            self.unit_system,
            series_count=5,
            vertical_ticks=8,
            y_axis_label=False,
        )
        self.box.addItems(
            [
                self.unit_system.format_name_unit(
                    self.chart.Y_OPTIONS[option].unit, name=option
                )
                for option in self.chart.Y_OPTIONS
            ]
        )

    def change_y(self, new_value):
        for option in self.chart.Y_OPTIONS:
            if option in new_value:
                self.y_option = option
        self.y_changed.emit()

    def update(self, *args, **kwargs):
        self.chart.update(*args, y_mode=self.y_option, **kwargs)



"""Contains a class holding current user settings."""
import dataclasses

from activate import serialise, units
from activate.app import connect, paths

DEFAULTS = {"unit_system": units.DEFAULT, "servers": [], "custom_units": {}}


def load_settings():
    """Load settings from a configuration file."""
    settings_data = DEFAULTS.copy()
    try:
        settings_data.update(serialise.load(paths.SETTINGS))
    except FileNotFoundError:
        pass

    settings_data["servers"] = [connect.Server(**s) for s in settings_data["servers"]]
    return Settings(**settings_data)


@dataclasses.dataclass
class Settings:
    """A settings configuration"""

    unit_system: str
    custom_units: dict
    servers: list

    def save(self):
        """Save settings to a configuration file."""
        serialise.dump(dataclasses.asdict(self), paths.SETTINGS, readable=True)



from PyQt5 import QtWidgets
from PyQt5.QtCore import Qt

Unchecked, PartiallyChecked, Checked = Qt.Unchecked, Qt.PartiallyChecked, Qt.Checked


class CheckList(QtWidgets.QListWidget):
    """A QListWidget with checkboxes on items."""

    def __init__(self, *args, **kwargs):
        self.do_not_recurse = False
        self.all_row = False
        super().__init__(*args, **kwargs)
        self.itemChanged.connect(self.item_changed)
        self.itemDoubleClicked.connect(self.item_double_clicked)

    def __getitem__(self, index):
        if isinstance(index, slice):
            return [x for x in self][index]
        result = self.item(index)
        if result is None:
            raise IndexError(f"{self.__class__.__qualname__} index out of range")
        return result

    @property
    def row_names(self):
        return [row.text() for row in self]

    @row_names.setter
    def row_names(self, new_items):
        self.clear()
        self.addItems(new_items)
        for row in self:
            row.setCheckState(Unchecked)

    @property
    def states(self):
        return {row.text(): row.checkState() for row in self}

    @states.setter
    def states(self, new_states):
        for index, item in enumerate(self.row_names):
            if item in new_states:
                self.set_check_state(index, new_states[item])

    def get_row(self, row):
        """Get a row from a string, index or row."""
        if isinstance(row, str):
            for real_row in self:
                if real_row.text() == row:
                    return real_row
            raise ValueError(f"{row} is not a row.")
        if isinstance(row, int):
            return self[row]
        return row

    def set_check_state(self, row, state):
        self.get_row(row).setCheckState(state)

    def check_state(self, row):
        return self.get_row(row).checkState()

    @property
    def checked_rows(self):
        return [r.text() for r in self if r.checkState() == Checked]

    def item_changed(self, item):
        if self.do_not_recurse or not self.all_row:
            return
        self.stop_updates()
        if self.is_all(item):
            for item_ in self[1:]:
                item_.setCheckState(item.checkState())
        else:
            states = set(i.checkState() for i in self[1:])
            self.set_all_state(
                next(iter(states)) if len(states) == 1 else PartiallyChecked
            )
        self.start_updates()

    def item_double_clicked(self, item):
        if self.is_all(item):
            self.set_all_state(Checked)
            return
        self.stop_updates()

        if self.all_row and len(self) > 2:
            self.set_all_state(PartiallyChecked)

        for item_ in self:
            if not self.is_all(item_):
                item_.setCheckState(Checked if item_ is item else Unchecked)
        self.start_updates()

    def check_all(self):
        for row in self:
            row.setCheckState(Checked)

    def add_all_row(self):
        self.insertItem(0, "All")
        self.all_row = True

    def is_all(self, item):
        """Check if a row is the 'All' row."""
        return self.all_row and self.row(item) == 0

    def set_all_state(self, state):
        if self.all_row:
            self.set_check_state(0, state)

    def stop_updates(self):
        self.do_not_recurse = True
        self.blockSignals(True)

    def start_updates(self):
        self.do_not_recurse = False
        self.blockSignals(False)



from collections import Counter
from datetime import datetime

from PyQt5 import QtWidgets

from activate import activity_types, number_formats
from activate.app import charts
from activate.app.dialogs import progress
from activate.app.ui.summary import Ui_summary

NOW = datetime.now()


class Summary(QtWidgets.QWidget, Ui_summary):
    """A widget summarising all of a person's activities."""

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.setupUi(self)

    def setup(self, unit_system, map_widget, activities):
        self.unit_system = unit_system
        self.map_widget = map_widget
        self.activities = activities
        # This has to be added here so when the heatmap is switched to,
        # the widget is already 'there', so it has a size. This lets
        # fitBounds work properly.
        self.heatmap_layout.addWidget(self.map_widget)
        self.records_table.set_units(self.unit_system)
        self.summary_period = "Year"
        self.progression_chart.set_units(self.unit_system)
        self.update_activity_types_list()
        self.eddington_chart = charts.LineChart(
            self.eddington_chart_widget, self.unit_system, series_count=2
        )
        self.eddington_chart.y_axis.setTitleText("Count")
        self.eddington_chart.add_legend(("Done", "Target"))

    def update_activity_types_list(self):
        """Set up activity types list for the summary page."""
        self.activity_types_list.row_names = [
            x[0] for x in Counter(a.sport for a in self.activities).most_common()
        ]
        self.activity_types_list.add_all_row()
        self.activity_types_list.check_all()

    def summary_tab_switch(self):
        tab = self.summary_tabs.tabText(self.summary_tabs.currentIndex())
        {
            "Totals": self.update_totals,
            "Records": self.update_records,
            "Progression": self.update_progression,
            "Gallery": self.update_gallery,
            "Heatmap": self.update_heatmap,
            "Eddington": self.update_eddington,
        }[tab]()

    def update_progression(self):
        """Update the progression chart."""
        self.progression_chart.update(
            self.summary_period,
            self.get_allowed_for_summary(),
            now=NOW,
            activities=self.activities,
        )

    def get_allowed_for_summary(self):
        """Get the allowed activity types from the checklist."""
        return set(self.activity_types_list.checked_rows)

    def set_formatted_number_label(self, label, value, dimension):
        """Set a label to a number, formatted with the correct units."""
        label.setText(
            number_formats.default_as_string(self.unit_system.format(value, dimension))
        )

    def update_totals(self):
        """Update the summary page totals."""
        allowed_activity_types = self.get_allowed_for_summary()
        allowed_activities = list(
            self.activities.filtered(allowed_activity_types, self.summary_period, NOW)
        )
        self.set_formatted_number_label(
            self.total_distance_label,
            self.activities.total_distance(allowed_activities),
            "distance",
        )
        self.set_formatted_number_label(
            self.total_time_label,
            self.activities.total_time(allowed_activities),
            "time",
        )
        self.total_activities_label.setText(
            str(self.activities.total_activities(allowed_activities))
        )
        self.set_formatted_number_label(
            self.total_climb_label,
            self.activities.total_climb(allowed_activities),
            "altitude",
        )

    def update_records(self):
        good_distances = {}
        for sport in self.get_allowed_for_summary():
            good_distances.update(
                activity_types.SPECIAL_DISTANCES[sport]
                if sport in activity_types.SPECIAL_DISTANCES
                else activity_types.SPECIAL_DISTANCES[None]
            )
        good_distances = {k: good_distances[k] for k in sorted(good_distances)}
        records, activity_ids = self.activities.get_records(
            self.get_allowed_for_summary(),
            self.summary_period,
            NOW,
            good_distances,
            lambda x: progress(self, list(x), "Loading"),
        )

        self.records_table.update_data(
            list(good_distances.values()), records, activity_ids
        )

    def update_gallery(self):
        self.gallery.replace_photos(
            self.activities.get_all_photos(
                self.get_allowed_for_summary(), self.summary_period, NOW
            )
        )

    def update_heatmap(self):
        self.heatmap_layout.addWidget(self.map_widget)
        self.map_widget.show_heatmap(
            self.activities.get_all_routes(
                self.get_allowed_for_summary(),
                self.summary_period,
                NOW,
                lambda x: progress(self, list(x), "Loading"),
            )
        )

    def update_eddington(self):
        allowed_activities = list(
            self.activities.filtered(
                self.get_allowed_for_summary(), self.summary_period, NOW
            )
        )
        if not allowed_activities:
            return
        unit = self.unit_system.units["distance"].size
        eddington_data = self.activities.eddington(allowed_activities, unit)
        eddington_number = 0
        for eddington_number in range(1, len(eddington_data) + 1):
            if eddington_data[eddington_number - 1] <= eddington_number * unit:
                break
        self.total_eddington_label.setText(
            f"{eddington_number} {self.unit_system.units['distance'].symbol}"
        )
        y_indices = list(range(1, len(eddington_data) + 1))
        x_indices = [x * unit for x in y_indices[: int(eddington_data[0] / unit) + 1]]
        y_indices = (y_indices, None)
        self.eddington_chart.update(
            (
                ((eddington_data, "distance"), y_indices),
                ((x_indices, "distance"), y_indices),
            )
        )

    def summary_period_changed(self, value):
        self.summary_period = value
        self.summary_tab_switch()



"""Load the fields from a GPX file on disk."""
from activate import times
from activate.filetypes import load_xml

CADENCE_FIELDS = ("./extensions/cadence", "./extensions/TrackPointExtension/cad")
HEART_RATE_FIELDS = (
    "./extensions/heartrate",
    "./extensions/hr",
    "./extensions/TrackPointExtension/hr",
)
FIELDS = {
    "lat": lambda p: float(p.get("lat")),
    "lon": lambda p: float(p.get("lon")),
    "ele": lambda p: float(p.find("./ele").text),
    "time": lambda p: times.from_GPX(p.find("./time").text),
    "speed": lambda p: float(p.find("./extensions/speed").text),
    "distance": lambda p: float(p.find("./extensions/distance").text),
    "cadence": lambda p: float(load_xml.try_multi(p, CADENCE_FIELDS)) / 60,
    "heartrate": lambda p: float(load_xml.try_multi(p, HEART_RATE_FIELDS)) / 60,
    "power": lambda p: float(p.find("./extensions/power").text),
}


def load_gpx(filename) -> tuple:
    """Extract the fields from a GPX file."""
    tree = load_xml.get_tree(filename)
    # Find the activity name
    try:
        name = tree.find("./trk/name").text
    except AttributeError:
        name = None

    # Find the sport
    try:
        sport = tree.find("./trk/type").text
        print(f"{sport=} {name=}")
    except AttributeError:
        sport = "unknown"

    points = tree.findall("./trk/trkseg/trkpt")
    return (name, sport, load_xml.load_fields(points, FIELDS))


def to_route(activity):
    name, track = activity.name, activity.track
    gpx = [
        '<?xml version="1.0" encoding="UTF-8"?><gpx version="1.0">'
        f"<name>{name}</name><trk><trkseg>"
    ]
    for p in range(len(track)):
        gpx.append(
            f'<trkpt lat="{track["lat"][p]}" lon="{track["lon"][p]}">'
            f'<ele>{track["ele"][p]}</ele></trkpt>"'
        )
    gpx.append("</trkseg></trk></gpx>")
    return "".join(gpx)



"""Classes inheriting from QTableWidgets or QTableWidgetItems."""
import dataclasses

from PyQt5 import QtCore, QtWidgets
from PyQt5.QtCore import Qt
from PyQt5.QtGui import QIcon

from activate import number_formats, times, units
from activate.app import connect


def aligned(horizontal="c", vertical="c"):
    """
    Get a Qt Alignment from a string.

    The first argument gives the horizontal alignment:
    Left, Centre, Right or Justified.
    The second argument gives the vertical alignment:
    Bottom, Centre, Top or bAseline.
    """
    return {
        "l": Qt.AlignLeft,
        "c": Qt.AlignHCenter,
        "r": Qt.AlignRight,
        "v": Qt.AlignJustify,
    }[horizontal.casefold()] | {
        "b": Qt.AlignBottom,
        "c": Qt.AlignVCenter,
        "t": Qt.AlignTop,
        "a": Qt.AlignBaseline,
    }[
        vertical.casefold()
    ]


def iterablise(obj):
    """If obj is not already iterable, form an endless iterator of it."""
    try:
        iter(obj)
        return obj
    except TypeError:
        # Infinite iterator
        return iter(lambda: obj, object())


def create_table_item(
    item, format_function=None, unit_system=None, align=None
) -> QtWidgets.QTableWidgetItem:
    """
    Create a table item that can be a FormattableNumber.

    If item is a tuple, will return a table item that looks like item[1]
    but sorts with item[0]. Otherwise just returns a normal table item.
    """
    if isinstance(item, units.DimensionValue):
        item = item.encode(unit_system)

    if format_function is not None:
        widget = FormattableNumber(item, format_function(item))
        widget.setTextAlignment(aligned("r"))
    # Format as string
    else:
        widget = QtWidgets.QTableWidgetItem(str(item))
    if align is not None:
        widget.setTextAlignment(align)
    return widget


def good_minus(string):
    """Replace an initial hyphen-minuses with a real minus sign."""
    if string and string[0] == "-":
        return "\u2212" + string[1:]
    return string


class FormattableNumber(QtWidgets.QTableWidgetItem):
    """A sortable, formatted number to place in a table."""

    def __init__(self, number, text):
        super().__init__(good_minus(text))
        self.number = number

    def __lt__(self, other):
        return self.number < other.number


class Table(QtWidgets.QTableWidget):
    unit_system = None

    def resize_to_contents(self, direction="h"):
        """
        Set a header to auto-resize its items.

        This also stops the user resizing them, which is good because
        usually resizing these things is not particularly useful.
        """
        if direction == "h":
            header = self.horizontalHeader()
        elif direction == "v":
            header = self.verticalHeader()
        else:
            raise ValueError(f"Invalid direction: {direction}. (Must be 'h' or 'v')")

        header.setSectionResizeMode(QtWidgets.QHeaderView.ResizeToContents)

    def set_item(self, row, column, value, format_function=None, align=None):
        self.setItem(
            row,
            column,
            create_table_item(value, format_function, self.unit_system, align=align),
        )

    def set_row(self, values, position, formats=None, alignments=None):
        alignments = iterablise(alignments)
        formats = iterablise(formats)
        for column, (value, format, align) in enumerate(
            zip(values, formats, alignments)
        ):
            self.set_item(position, column, value, format, align)

    def get_row_text(self, index) -> list:
        result = []
        for column in range(self.columnCount()):
            item = self.item(index, column)
            result.append(None if item is None else item.text())
        return result

    @property
    def headings(self):
        return [self.get_heading(i) for i in range(self.columnCount())]

    def get_heading(self, index) -> str:
        return self.horizontalHeaderItem(index).text()


class ValueColumnTable(QtWidgets.QTableWidget):
    def set_units(self, unit_system):
        self.unit_system = unit_system

    def resize_to_contents(self, vertical=False):
        if vertical:
            header = self.verticalHeader()
        else:
            header = self.horizontalHeader()
        header.setSectionResizeMode(QtWidgets.QHeaderView.ResizeToContents)

    def define_columns(self, names, dimensions, format_functions=None, alignments=None):
        self.setColumnCount(len(names))
        self.setHorizontalHeaderLabels(
            [
                self.unit_system.format_name_unit(
                    name, self.unit_system.units[dimension].symbol
                )
                for name, dimension in zip(names, dimensions)
            ]
        )
        if format_functions is None:
            self.format_functions = [None for _ in names]
        else:
            self.format_functions = format_functions
        if alignments is None:
            self.alignments = [None for _ in names]
        else:
            self.alignments = alignments

    def set_item(self, row, column, value):
        self.setItem(
            row,
            column,
            create_table_item(
                value,
                self.format_functions[column],
                self.unit_system,
                self.alignments[column],
            ),
        )

    def set_row(self, values, position):
        for column, value in enumerate(values):
            self.set_item(position, column, value)

    def __len__(self):
        return self.rowCount()


class SplitTable(ValueColumnTable):
    headings = ["Number", "Time", "Split", "Speed", "Net Climb", "Ascent"]
    dimensions = [None, "time", "time", "speed", "altitude", "altitude"]

    def set_units(self, *args, **kwargs):
        super().set_units(*args, **kwargs)
        self.define_columns(
            self.headings,
            self.dimensions,
            [number_formats.split_format(h) for h in self.headings],
            alignments=[aligned("r") for _ in self.headings],
        )

        self.resize_to_contents()

    def update_data(self, data):
        self.setRowCount(len(data))
        for y, row in enumerate(data):
            row_data = [y + 1] + row
            self.set_row(row_data, y)


class ActivityListTable(ValueColumnTable):
    headings = ["Name", "Type", "Start Time", "Distance"]
    dimensions = [None, None, None, "distance"]
    right_clicked = QtCore.pyqtSignal(QtCore.QEvent)

    def set_units(self, *args, **kwargs):
        super().set_units(*args, **kwargs)
        self.define_columns(
            self.headings,
            self.dimensions,
            [number_formats.list_format(h) for h in self.headings],
        )
        self.resize_to_contents()
        self.resize_to_contents(vertical=True)
        self.horizontalHeader().setSectionResizeMode(0, QtWidgets.QHeaderView.Stretch)

    @property
    def current_activity_id(self):
        return self.selectedItems()[0].activity_id

    def set_id_row(self, activity_id, values, position):
        """
        Set the items in the given activity list row to specific values.

        Assigns activity_id to the item in column zero in the new_row.
        """
        self.set_row(values, position)
        self.item(position, 0).activity_id = activity_id

    def add_id_row(self, activity_id, values, position):
        sorting_was_enabled = self.isSortingEnabled()
        self.setSortingEnabled(False)
        self.insertRow(position)
        self.set_id_row(activity_id, values, position)
        if sorting_was_enabled:
            self.setSortingEnabled(True)

    def default_sort(self):
        self.setSortingEnabled(True)
        self.sortItems(self.headings.index("Start Time"), Qt.DescendingOrder)

    def mouseReleaseEvent(self, event):
        if event.button() == Qt.RightButton:
            self.right_clicked.emit(event)
        else:
            super().mouseReleaseEvent(event)

    def select(self, activity_id):
        for row in range(len(self)):
            if self.item(row, 0).activity_id == activity_id:
                self.selectRow(row)
                return

    def filter(self, ids):
        for row in range(len(self)):
            self.setRowHidden(row, self.item(row, 0).activity_id not in ids)


class SocialActivityList(ActivityListTable):
    headings = ["Server", "User", "Name", "Type", "Start Time", "Distance"]
    dimensions = [None, None, None, None, None, "distance"]

    def filter_by_server(self, allowed):
        for row in range(len(self)):
            servers = self.item(row, 0).text().split("\n")
            usernames = self.item(row, 1).text().split("\n")
            self.setRowHidden(
                row, not any(x in allowed for x in zip(servers, usernames))
            )


class CurveTable(ValueColumnTable):
    headings = ["Distance", "Time", "Speed"]
    dimensions = [None, "time", "speed"]

    def set_units(self, *args, **kwargs):
        super().set_units(*args, **kwargs)
        self.define_columns(
            self.headings,
            self.dimensions,
            [lambda x: x, times.to_string, lambda x: str(round(x, 1))],
        )

    def update_data(self, good_distance_names, table):
        self.setRowCount(len(table))
        for index, row in enumerate(table):
            self.set_row(good_distance_names[index : index + 1] + list(row[1:]), index)


class RecordsTable(CurveTable):
    headings = CurveTable.headings + ["Activity"]
    dimensions = CurveTable.dimensions + [None]
    # An int results in an overflow issue
    gone_to = QtCore.pyqtSignal(object)

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.itemDoubleClicked.connect(self.go_to)

    def set_units(self, *args, **kwargs):
        super().set_units(*args, **kwargs)
        self.define_columns(
            self.headings,
            self.dimensions,
            [lambda x: x, times.to_string, lambda x: str(round(x, 1)), lambda x: x],
        )
        self.resize_to_contents()

    def update_data(self, good_distance_names, table, activity_ids):
        super().update_data(good_distance_names, table)
        for row, activity_id in enumerate(activity_ids):
            self.item(row, self.headings.index("Activity")).activity_id = activity_id

    def go_to(self, item):
        if self.headings[item.column()] == "Activity":
            self.gone_to.emit(item.activity_id)


class InfoTable(Table):
    """
    The table widget on the right.

    This is used for distance, climb, duration etc.
    """

    def set_units(self, unit_system):
        self.unit_system = unit_system

    def update_data(self, info: dict):
        self.setRowCount(len(info))
        for row, (field, value) in enumerate(info.items()):
            self.set_item(row, 0, field)
            self.set_item(
                row, 1, value, number_formats.info_format(field), align=aligned("r")
            )
            self.set_item(
                row,
                2,
                self.unit_system.units[value.dimension].symbol,
                align=aligned("l"),
            )


class ServersTable(Table):
    headings = ["Address", "Name", "Username", "Password"]

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        # Set up right click delete menu
        self.menu = QtWidgets.QMenu(self)
        self.action_delete = QtWidgets.QAction("Delete")
        self.action_delete.setIcon(QIcon.fromTheme("edit-delete"))
        self.menu.addAction(self.action_delete)

    def set_columns(self):
        self.setColumnCount(len(self.headings))
        self.setHorizontalHeaderLabels(self.headings)

    def set_server(self, row, server):
        self.set_row(["" if x is None else x for x in dataclasses.astuple(server)], row)

    def set_servers(self, servers):
        self.setRowCount(len(servers))
        self.set_columns()
        for row, server in enumerate(servers):
            self.set_server(row, server)

    def get_servers(self):
        return [connect.Server(*self.get_row_text(r)) for r in range(self.rowCount())]

    def add_row(self):
        self.setRowCount(self.rowCount() + 1)
        self.set_columns()

    def contextMenuEvent(self, event):
        """Open a context menu to remove a server."""
        if self.menu.exec(event.globalPos()):
            self.removeRow(self.row(self.itemAt(event.pos())))



"""Define some standard paths for storing data."""
from pathlib import Path

from PyQt5.QtCore import QStandardPaths

# Locations for user data
DATA = (
    Path(QStandardPaths.writableLocation(QStandardPaths.AppDataLocation)) / "activate"
)
ACTIVITIES = DATA / "activities"
TRACKS = DATA / "originals"
PHOTOS = DATA / "photos"

# Location for configuration files
SETTINGS = (
    Path(QStandardPaths.writableLocation(QStandardPaths.ConfigLocation))
    / "activate.json"
)

HOME = Path(QStandardPaths.writableLocation(QStandardPaths.HomeLocation))


def ensure_all_present():
    """Create all the required directories."""
    for path in (ACTIVITIES, TRACKS, PHOTOS):
        path.mkdir(parents=True, exist_ok=True)



# -*- coding: utf-8 -*-

# Form implementation generated from reading ui file 'resources/ui/activity_summary.ui'
#
# Created by: PyQt5 UI code generator 5.15.2
#
# WARNING: Any manual changes made to this file will be lost when pyuic5 is
# run again.  Do not edit this file unless you know what you are doing.

from PyQt5 import QtCore, QtGui, QtWidgets


class Ui_activity_summary(object):
    def setupUi(self, activity_summary):
        activity_summary.setObjectName("activity_summary")
        activity_summary.resize(634, 300)
        self.activity_summary_layout = QtWidgets.QGridLayout(activity_summary)
        self.activity_summary_layout.setObjectName("activity_summary_layout")
        self.date_activity_type_area = QtWidgets.QHBoxLayout()
        self.date_activity_type_area.setObjectName("date_activity_type_area")
        self.date_time_label = QtWidgets.QLabel(activity_summary)
        font = QtGui.QFont()
        font.setPointSize(11)
        font.setBold(True)
        font.setWeight(75)
        self.date_time_label.setFont(font)
        self.date_time_label.setObjectName("date_time_label")
        self.date_activity_type_area.addWidget(self.date_time_label)
        self.activity_type_label = QtWidgets.QLabel(activity_summary)
        self.activity_type_label.setMaximumSize(QtCore.QSize(120, 16777215))
        font = QtGui.QFont()
        font.setPointSize(11)
        font.setBold(True)
        font.setWeight(75)
        self.activity_type_label.setFont(font)
        self.activity_type_label.setObjectName("activity_type_label")
        self.date_activity_type_area.addWidget(self.activity_type_label)
        self.activity_summary_layout.addLayout(self.date_activity_type_area, 0, 1, 1, 1)
        self.info_table = InfoTable(activity_summary)
        self.info_table.setMinimumSize(QtCore.QSize(310, 0))
        self.info_table.setMaximumSize(QtCore.QSize(310, 16777215))
        self.info_table.setHorizontalScrollBarPolicy(QtCore.Qt.ScrollBarAlwaysOff)
        self.info_table.setEditTriggers(QtWidgets.QAbstractItemView.NoEditTriggers)
        self.info_table.setSelectionMode(QtWidgets.QAbstractItemView.SingleSelection)
        self.info_table.setShowGrid(False)
        self.info_table.setColumnCount(3)
        self.info_table.setObjectName("info_table")
        self.info_table.setRowCount(0)
        self.info_table.horizontalHeader().setVisible(False)
        self.info_table.horizontalHeader().setDefaultSectionSize(100)
        self.info_table.horizontalHeader().setStretchLastSection(True)
        self.info_table.verticalHeader().setVisible(False)
        self.activity_summary_layout.addWidget(self.info_table, 2, 1, 1, 1)
        self.activity_text_area = QtWidgets.QVBoxLayout()
        self.activity_text_area.setObjectName("activity_text_area")
        self.activity_name_and_flags = QtWidgets.QHBoxLayout()
        self.activity_name_and_flags.setObjectName("activity_name_and_flags")
        self.activity_name_label = QtWidgets.QLabel(activity_summary)
        font = QtGui.QFont()
        font.setPointSize(20)
        font.setBold(True)
        font.setWeight(75)
        self.activity_name_label.setFont(font)
        self.activity_name_label.setObjectName("activity_name_label")
        self.activity_name_and_flags.addWidget(self.activity_name_label)
        self.flags_label = QtWidgets.QLabel(activity_summary)
        font = QtGui.QFont()
        font.setPointSize(11)
        self.flags_label.setFont(font)
        self.flags_label.setAlignment(
            QtCore.Qt.AlignRight | QtCore.Qt.AlignTrailing | QtCore.Qt.AlignVCenter
        )
        self.flags_label.setIndent(20)
        self.flags_label.setObjectName("flags_label")
        self.activity_name_and_flags.addWidget(self.flags_label)
        self.activity_text_area.addLayout(self.activity_name_and_flags)
        self.description_label = QtWidgets.QLabel(activity_summary)
        self.description_label.setWordWrap(True)
        self.description_label.setObjectName("description_label")
        self.activity_text_area.addWidget(self.description_label)
        self.activity_summary_layout.addLayout(self.activity_text_area, 0, 0, 1, 1)
        self.map_container = QtWidgets.QGridLayout()
        self.map_container.setObjectName("map_container")
        self.activity_summary_layout.addLayout(self.map_container, 1, 0, 2, 1)

        self.retranslateUi(activity_summary)
        QtCore.QMetaObject.connectSlotsByName(activity_summary)

    def retranslateUi(self, activity_summary):
        _translate = QtCore.QCoreApplication.translate
        activity_summary.setWindowTitle(_translate("activity_summary", "Form"))
        self.date_time_label.setText(_translate("activity_summary", "[Date]"))
        self.activity_type_label.setText(_translate("activity_summary", "[Type]"))
        self.activity_name_label.setText(
            _translate("activity_summary", "[Activity Name]")
        )
        self.flags_label.setText(_translate("activity_summary", "[Flags]"))
        self.description_label.setText(_translate("activity_summary", "[Description]"))


from activate.app.tables import InfoTable



"""
Functions for serialising and deserialising objects to JSON.

The JSON can optionally be gzipped.
"""
import gzip
import json
import uuid
from datetime import datetime, timedelta
from pathlib import Path


def default(obj):
    """
    Turn datetimes and timedeltas into JSON.

    >>> default(datetime(2000, 1, 2, 12, 30, 42))
    {'__DATETIME': '2000-01-02T12:30:42'}
    >>> default(timedelta(minutes=1, seconds=40))
    {'__TIMEDELTA': 100.0}
    """
    if isinstance(obj, datetime):
        return {"__DATETIME": obj.isoformat()}
    if isinstance(obj, timedelta):
        return {"__TIMEDELTA": obj.total_seconds()}
    if isinstance(obj, uuid.UUID):
        return {"__ID": str(obj)}
    if isinstance(obj, Path):
        return {"__PATH": str(obj)}
    raise TypeError(f"Cannot serialise {obj.__class__.__qualname__}")


DECODE_KEYS = {
    "__DATETIME": datetime.fromisoformat,
    "__TIMEDELTA": lambda v: timedelta(seconds=v),
    "__ID": uuid.UUID,
    "__PATH": Path,
}


def decode(obj):
    """Undo encoding done by default."""
    if len(obj) != 1:
        return obj
    key, value = obj.popitem()
    try:
        return DECODE_KEYS[key](value)
    except KeyError:
        return {key: value}


def dumps(obj, readable=False):
    """Convert an object to a JSON string."""
    return json.dumps(
        obj,
        default=default,
        separators=None if readable else (",", ":"),
        indent="\t" if readable else None,
    )


def dump_bytes(obj, gz=False, readable=False):
    """Convert an object to data."""
    data = dumps(obj, readable=readable).encode("utf-8")
    return gzip.compress(data) if gz else data


def loads(data, gz=False):
    """Load an object from a string or bytes."""
    data = gzip.decompress(data) if gz else data
    return json.loads(data, object_hook=decode)


def dump(obj, filename, *args, **kwargs):
    """
    Save obj as a JSON file. Can store datetimes and timedeltas.

    Can be gzipped if gz is True.
    """
    with open(filename, "wb") as f:
        f.write(dump_bytes(obj, *args, **kwargs))


def load(filename: Path, gz="auto"):
    """Load a JSON file. Can retrieve datetimes and timedeltas."""
    if gz == "auto":
        gz = filename.suffix.casefold() == ".gz"
    with (gzip.open if gz else open)(filename, "rt") as f:
        return json.load(f, object_hook=decode)



import hashlib
import json
import sqlite3
from base64 import b64decode, b64encode
from datetime import timedelta
from functools import wraps
from pathlib import Path
from uuid import UUID

import pkg_resources
from flask import Flask, abort, g, request

from activate import activity, serialise

DATA_DIR = Path("/var/lib/activate")

USERS_FILE = DATA_DIR / "users.json"

app = Flask(__name__)

ACTIVITIES_DIR = DATA_DIR / "activities"
ACTIVITIES_DIR.mkdir(parents=True, exist_ok=True)

ACTIVITIES_DATABASE_PATH = DATA_DIR / "activities.sqlite"

sqlite3.register_converter("DICT", serialise.loads)
sqlite3.register_adapter(dict, serialise.dumps)

sqlite3.register_converter(
    "TIMEDELTA", lambda d: timedelta(seconds=float(d) if b"." in d else int(d))
)
sqlite3.register_adapter(timedelta, lambda d: d.total_seconds())

sqlite3.register_converter("UUID", lambda u: UUID(u.decode("utf-8")))
sqlite3.register_adapter(UUID, str)


def add_row(database, table: str, values: dict):
    """Add a row to an SQLite database."""
    database.execute(
        (
            f"INSERT INTO {table}"
            f" ({', '.join(values)})"
            f" VALUES ({', '.join(f':{v}' for v in values)})"
        ),
        values,
    )


def get_row(database, table: str, values: dict):
    """Find a row in an SQLite database."""
    return database.execute(
        f"SELECT * FROM {table} WHERE {' AND '.join(f'{v} = :{v}' for v in values)}",
        values,
    ).fetchone()


def delete_by_id(activities, activity_id):
    """Delete a row with a given activity_id."""
    activities.execute(
        "DELETE FROM activities WHERE activity_id = ?", [str(activity_id)]
    )


def reset_activities():
    """Generate a blank activities database."""
    db = load_database()
    db.executescript(
        pkg_resources.resource_string("activate.resources", "init.sql").decode("utf-8")
    )
    db.commit()


def load_database():
    db = sqlite3.connect(
        ACTIVITIES_DATABASE_PATH,
        detect_types=sqlite3.PARSE_DECLTYPES | sqlite3.PARSE_COLNAMES,
    )
    db.row_factory = sqlite3.Row
    return db


def get_activities():
    if "db" not in g:
        g.db = load_database()

    return g.db


@app.teardown_appcontext
def close_activities(e=None):
    db = g.pop("db", None)

    if db is not None:
        db.close()


def get_users():
    try:
        with open(USERS_FILE) as f:
            return json.load(f)
    except FileNotFoundError:
        return {}


def save_users(users):
    with open(USERS_FILE, "w") as f:
        json.dump(f, users)


def password_hash(password: str, salt):
    return b64encode(
        hashlib.scrypt(
            password.encode("utf-8"), salt=b64decode(salt), n=16384, r=8, p=1
        )
    ).decode("utf-8")


def verify_request():
    """Check username and password against the database."""
    if request.authorization is None:
        return False
    username = request.authorization["username"]
    if username not in users:
        return False
    password = request.authorization["password"]
    return (
        password_hash(password, users[username]["salt"])
        == users[username]["password_hash"]
    )


def requires_auth(function):
    @wraps(function)
    def new_function(*args, **kwargs):
        if not verify_request():
            abort(403)
        return function(*args, **kwargs)

    return new_function


@app.route("/")
def index():
    return "This is an Activate server."


@app.route("/api/send_activity", methods=["POST"])
@requires_auth
def upload():
    data = serialise.loads(request.form["activity"])
    data["username"] = request.authorization["username"]
    new_activity = activity.Activity(**data)
    activities = get_activities()
    delete_by_id(activities, new_activity.activity_id)
    add_row(
        activities,
        "activities",
        {
            "name": new_activity.name,
            "sport": new_activity.sport,
            "flags": new_activity.flags,
            "start_time": new_activity.start_time,
            "distance": new_activity.distance,
            "duration": new_activity.track.elapsed_time,
            "climb": new_activity.track.ascent,
            "activity_id": new_activity.activity_id,
            "username": new_activity.username,
        },
    )
    new_activity.save(ACTIVITIES_DIR)
    activities.commit()
    return "DONE"


@app.route("/api/delete_activity/<string:activity_id>")
@requires_auth
def delete_activity(activity_id):
    activity_id = UUID(activity_id)
    activities = get_activities()
    row = get_row(activities, "activities", {"activity_id": activity_id})
    if row["username"] != request.authorization["username"]:
        abort(403)
    delete_by_id(activities, activity_id)
    activities.commit()
    return "DONE"


@app.route("/api/get_activities")
@requires_auth
def get_list():
    activities = get_activities()
    activities = activities.execute("SELECT * FROM activities").fetchall()
    return serialise.dump_bytes([{k: a[k] for k in a.keys()} for a in activities])


@app.route("/api/get_activity/<string:activity_id>")
@requires_auth
def get_activity(activity_id):
    activity_id = UUID(activity_id)
    activities = get_activities()
    try:
        get_row(activities, "activities", {"activity_id": activity_id})
    except FileNotFoundError:
        abort(404)
    with open(ACTIVITIES_DIR / f"{activity_id}.json.gz", "rb") as f:
        return f.read()


users = get_users()
if not ACTIVITIES_DATABASE_PATH.exists():
    reset_activities()



import shutil
from uuid import uuid4

from activate import serialise
from activate import track as track_
from activate.units import DimensionValue


def from_track(name, sport, track, filename):
    return Activity(name, sport, track, filename)


def none_default(value, default):
    """Return default if value is None else value."""
    return default if value is None else value


class Activity:
    def __init__(
        self,
        name,
        sport,
        track,
        original_name,
        flags=None,
        start_time=None,
        distance=None,
        activity_id=None,
        description="",
        photos=None,
        server=None,
        username=None,
    ):
        self.name = name
        self.sport = sport
        if isinstance(track, dict):
            self.track = track_.Track(track)
        else:
            self.track = track
        self.original_name = original_name
        self.server = server
        self.username = username
        self.flags = none_default(flags, {})
        self.start_time = none_default(start_time, self.track.start_time)
        self.distance = none_default(distance, self.track.length)
        self.activity_id = none_default(activity_id, uuid4())
        self.description = description
        self.photos = none_default(photos, [])

    @property
    def stats(self):
        result = {}
        result["Distance"] = DimensionValue(self.distance, "distance")
        result["Elapsed Time"] = DimensionValue(self.track.elapsed_time, "time")
        if self.track.moving_time < self.track.elapsed_time:
            result["Moving Time"] = DimensionValue(self.track.moving_time, "time")
        if self.track.has_altitude_data:
            result["Ascent"] = DimensionValue(self.track.ascent, "altitude")
            result["Descent"] = DimensionValue(self.track.descent, "altitude")
        average_speed = self.track.average("speed")
        if average_speed > 0:
            average_speed_moving = self.track.average_speed_moving
            if average_speed_moving / average_speed < 1.01:
                average_speed_moving = None
            result["Average Speed"] = DimensionValue(average_speed, "speed")
            if average_speed_moving is not None:
                result["Mov. Av. Speed"] = DimensionValue(average_speed_moving, "speed")
            result["Pace"] = DimensionValue(1 / average_speed, "pace")
            if average_speed_moving is not None:
                result["Pace (mov.)"] = DimensionValue(1 / average_speed_moving, "pace")
        if not self.track.manual:
            result["Max. Speed"] = DimensionValue(self.track.maximum("speed"), "speed")
        if self.track.has_altitude_data:
            result["Highest Point"] = DimensionValue(
                self.track.maximum("ele"), "altitude"
            )
        if "heartrate" in self.track:
            result["Average HR"] = DimensionValue(
                self.track.average("heartrate"), "heartrate"
            )
        if "cadence" in self.track:
            result["Avg. Cadence"] = DimensionValue(
                self.track.average("cadence"), "cadence"
            )
        if "power" in self.track:
            result["Average Power"] = DimensionValue(
                self.track.average("power"), "power"
            )
            result["Max. Power"] = DimensionValue(self.track.maximum("power"), "power")
        return result

    @property
    def active_flags(self):
        return [k for k, v in self.flags.items() if v]

    def unload(self, unloaded_class):
        return unloaded_class(
            self.name,
            self.sport,
            self.flags,
            self.start_time,
            self.distance,
            self.track.elapsed_time,
            self.track.ascent,
            self.activity_id,
            self.server,
            self.username,
        )

    @property
    def save_data(self):
        return {
            "name": self.name,
            "sport": self.sport,
            "track": self.track.save_data,
            "original_name": self.original_name,
            "flags": self.flags,
            "start_time": self.start_time,
            "distance": self.distance,
            "activity_id": self.activity_id,
            "description": self.description,
            "photos": self.photos,
            "server": self.server,
            "username": self.username,
        }

    def save(self, path):
        serialise.dump(self.save_data, path / f"{self.activity_id}.json.gz", gz=True)

    def export_original(self, filename):
        shutil.copy2(self.original_name, filename)



import collections
from datetime import timedelta

from PyQt5 import QtWidgets

from activate import activity_types, times, units
from activate.app import checklist

UNIVERSAL_FLAGS = ("Commute", "Indoor")
TYPE_FLAGS = collections.defaultdict(tuple)
TYPE_FLAGS.update(activity_types.FLAGS)


class ActivityFlagEdit(checklist.CheckList):
    def change_options(self, activity_type):
        self.row_names = TYPE_FLAGS[activity_type] + UNIVERSAL_FLAGS


class DurationEdit(QtWidgets.QFormLayout):
    """A widget to allow editing an hours minute seconds duration."""

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.hours_widget = QtWidgets.QSpinBox()
        self.addRow("Hours", self.hours_widget)
        self.minutes_widget = QtWidgets.QSpinBox()
        self.minutes_widget.setRange(0, 59)
        self.addRow("Minutes", self.minutes_widget)
        self.seconds_widget = QtWidgets.QDoubleSpinBox()
        self.seconds_widget.setRange(0, 59.99)
        self.addRow("Seconds", self.seconds_widget)

    def value(self):
        return timedelta(
            hours=self.hours_widget.value(),
            minutes=self.minutes_widget.value(),
            seconds=self.seconds_widget.value(),
        )

    def set_value(self, new: timedelta):
        hours, minutes, seconds = times.hours_minutes_seconds(new)
        self.hours_widget.setValue(hours)
        self.minutes_widget.setValue(minutes)
        self.seconds_widget.setValue(seconds)


WIDGET_VALUES = {
    QtWidgets.QLineEdit: lambda w: w.text(),
    QtWidgets.QPlainTextEdit: lambda w: w.toPlainText(),
    QtWidgets.QTextEdit: lambda w: w.text(),
    QtWidgets.QSpinBox: lambda w: w.value(),
    QtWidgets.QDoubleSpinBox: lambda w: w.value(),
    QtWidgets.QComboBox: lambda w: w.currentText(),
    QtWidgets.QTimeEdit: lambda w: w.time().toPyTime(),
    QtWidgets.QDateTimeEdit: lambda w: w.dateTime().toPyDateTime(),
    QtWidgets.QDateEdit: lambda w: w.date().toPyDate(),
    QtWidgets.QAbstractSlider: lambda w: w.value(),
    QtWidgets.QKeySequenceEdit: lambda w: w.keySequence(),
    checklist.CheckList: lambda w: w.states,
    DurationEdit: lambda w: w.value(),
}

WIDGET_SETTERS = {
    QtWidgets.QLineEdit: lambda w, v: w.setText(v),
    QtWidgets.QPlainTextEdit: lambda w, v: w.setPlainText(v),
    QtWidgets.QTextEdit: lambda w, v: w.setText(v),
    QtWidgets.QSpinBox: lambda w, v: w.setValue(v),
    QtWidgets.QDoubleSpinBox: lambda w, v: w.setValue(v),
    QtWidgets.QComboBox: lambda w, v: w.setCurrentText(v),
    QtWidgets.QTimeEdit: lambda w, v: w.setTime(v),
    QtWidgets.QDateTimeEdit: lambda w, v: w.setDateTime(v),
    QtWidgets.QDateEdit: lambda w, v: w.setDate(v),
    QtWidgets.QAbstractSlider: lambda w, v: w.setValue(v),
    QtWidgets.QKeySequenceEdit: lambda w, v: w.setKeySequence(v),
    checklist.CheckList: lambda w, v: setattr(w, "state", v),
    DurationEdit: lambda w, v: w.set_value(v),
}


def get_value(widget):
    for widget_type, function in WIDGET_VALUES.items():
        if isinstance(widget, widget_type):
            return function(widget)


def set_value(widget, value):
    for widget_type, function in WIDGET_SETTERS.items():
        if isinstance(widget, widget_type):
            return function(widget, value)


class Form(QtWidgets.QFormLayout):
    def __init__(self, fields, *args, **kwargs):
        self.fields = fields
        super().__init__(*args, **kwargs)
        for (name, entry) in self.fields.items():
            self.addRow(name, entry)

    def values(self):
        return {
            name: get_value(
                self.itemAt(index, self.FieldRole).widget()
                or self.itemAt(index, self.FieldRole).layout()
            )
            for index, name in enumerate(self.fields.keys())
        }

    def set_values(self, values):
        for index, name in enumerate(self.fields.keys()):
            if name in values:
                set_value(
                    self.itemAt(index, self.FieldRole).widget()
                    or self.itemAt(index, self.FieldRole).layout(),
                    values[name],
                )


class CustomUnits(QtWidgets.QFormLayout):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.combo_boxes = {}
        self.labels = list(units.ALL.keys())
        for unit, options in units.ALL.items():
            if len(options) <= 1:
                continue
            combo_box = QtWidgets.QComboBox()
            combo_box.addItem("Default")
            for option in options:
                combo_box.addItem(option.name.capitalize())
            self.addRow(unit.replace("_", " ").title(), combo_box)
            self.combo_boxes[unit] = combo_box

    def units_dict(self):
        result = {}
        for unit in self.combo_boxes:
            value = self.combo_boxes[unit].currentText()
            if value == "Default":
                continue
            result[unit] = value.casefold()
        return result

    def set_units(self, units):
        for unit, value in units.items():
            self.combo_boxes[unit].setCurrentIndex(
                self.combo_boxes[unit].findText(value.capitalize())
            )



"""Load the fields from a GPX file on disk."""
from activate import times
from activate.filetypes import load_xml

FIELDS = {
    "lat": lambda p: float(p.find("./Position/LatitudeDegrees").text),
    "lon": lambda p: float(p.find("./Position/LongitudeDegrees").text),
    "ele": lambda p: float(p.find("./AltitudeMeters").text),
    "time": lambda p: times.from_GPX(p.find("./Time").text),
    "speed": lambda p: float(p.find("./Extensions/TPX/Speed").text),
    "distance": lambda p: float(p.find("./DistanceMeters").text),
    "cadence": lambda p: float(p.find("./Extensions/TPX/RunCadence").text) / 60,
    "heartrate": lambda p: float(p.find("./HeartRateBpm").text) / 60,
    "power": lambda p: float(p.find("./Extensions/TPX/Watts").text),
}


def load_tcx(filename):
    """Extract the fields from a TCX file."""
    tree = load_xml.get_tree(filename)
    # Find the sport
    try:
        sport = tree.find("./Activities/Activity").get("Sport")
        print(f"{sport=} {filename=}")
    except AttributeError:
        sport = "unknown"

    points = tree.findall("./Activities/Activity/Lap/Track/Trackpoint")
    return (None, sport, load_xml.load_fields(points, FIELDS))



import pyqtlet
from PyQt5.QtCore import Qt
from pyqtlet import L


def get_bounds(*routes):
    """Find the area of the map."""
    return [
        [min(p[0] for r in routes for p in r), min(p[1] for r in routes for p in r)],
        [max(p[0] for r in routes for p in r), max(p[1] for r in routes for p in r)],
    ]


DEFAULT_POS = [53, -1]

ACTIVATE_COLOUR = "#802090"


def call_js_method(obj, method, *params) -> None:
    """Call obj.method(*params) equivalent in JavaScript."""
    obj.runJavaScript(f"{obj.jsName}.{method}({','.join(str(x) for x in params)});")


class Polyline(L.polyline):
    def setLatLngs(self, coordinates):
        call_js_method(self, "setLatLngs", coordinates)


class CircleMarker(L.circleMarker):
    def setLatLng(self, coordinates):
        call_js_method(self, "setLatLng", coordinates)


class Map(pyqtlet.MapWidget):
    def __init__(self, parent):
        super().__init__()
        size_policy = self.sizePolicy()
        size_policy.setRetainSizeWhenHidden(True)
        self.setSizePolicy(size_policy)
        self.setContextMenuPolicy(Qt.NoContextMenu)
        self.map = L.map(self)
        L.tileLayer(
            "http://{s}.tile.osm.org/{z}/{x}/{y}.png",
            {"attribution": "&copy; OpenStreetMap contributors"},
        ).addTo(self.map)

        self.map.runJavaScript(f"{self.map.jsName}.attributionControl.setPrefix('');")


class MapWidget(Map):
    """A map for displaying a route or heatmap"""

    def __init__(self, parent):
        super().__init__(parent)
        self.route_lines = []
        self.start_icon = CircleMarker(DEFAULT_POS, {"radius": 8, "color": "#10b020"})
        self.finish_icon = CircleMarker(DEFAULT_POS, {"radius": 8, "color": "#e00000"})
        self.marker = CircleMarker(DEFAULT_POS, {"radius": 5, "color": ACTIVATE_COLOUR})
        self.highlight_section = self.add_route_line(self.highlight_colour)
        self.mode = None

    def show_route(self, route: list):
        """Display a list of points on the map."""
        self.map.fitBounds(get_bounds(route))
        if self.mode != "route":
            self.clear_route_lines()
            self.route_lines = [self.add_route_line()]
        self.route_lines[0].setLatLngs(route)
        self.start_icon.setLatLng(route[0])
        self.finish_icon.setLatLng(route[-1])
        self.start_icon.addTo(self.map)
        self.finish_icon.addTo(self.map)
        self.mode = "route"

    def show_heatmap(self, routes: list):
        """Display lists of points on the map as a heatmap."""
        if not routes:
            return
        colour = ACTIVATE_COLOUR + hex(min(round(1000 / (len(routes) ** 0.5)), 255))[2:]
        self.map.fitBounds(get_bounds(*routes))
        self.start_icon.removeFrom(self.map)
        self.finish_icon.removeFrom(self.map)
        self.clear_route_lines()
        self.route_lines = []
        for route in routes:
            self.route_lines.append(self.add_route_line(colour))
            self.route_lines[-1].setLatLngs(route)
        self.mode = "heatmap"

    def add_route_line(self, colour=ACTIVATE_COLOUR):
        line = Polyline([], {"smoothFactor": 0, "color": colour})
        line.addTo(self.map)
        return line

    def clear_route_lines(self):
        while self.route_lines:
            self.route_lines.pop().removeFrom(self.map)

    def show_marker(self, position):
        if position is None:
            self.remove_marker()
            return
        self.marker.setLatLng(list(position))
        self.marker.addTo(self.map)

    def remove_marker(self):
        self.marker.removeFrom(self.map)

    def show_highlight(self, part):
        self.highlight_section.setLatLngs(part)
        self.highlight_section.addTo(self.map)
        call_js_method(self.highlight_section, "bringToFront")

    def remove_highlight(self):
        self.highlight_section.removeFrom(self.map)

    @property
    def highlight_colour(self):
        return self.palette().color(self.palette().Highlight).name()



import math
from datetime import timedelta

from activate import times


def as_int(value) -> str:
    """Format a value as an integer, ready to display."""
    return str(round(value))


def maybe_as_int(value) -> str:
    """
    Format a value as an integer if it is one.

    Any very close value will be formatted as an integer to avoid
    floating-point errors.
    """
    if math.isclose(value, round(value)):
        return as_int(value)
    return str(value)


def info_format(entry: str):
    """Format an value for the info box."""
    if entry in {"Average Speed", "Mov. Av. Speed", "Distance"}:
        return lambda value: f"{value:.2f}"
    if entry in {"Max. Speed", "Average Power", "Average HR", "Avg. Cadence"}:
        return lambda value: f"{value:.1f}"
    if entry in {"Ascent", "Descent", "Highest Point", "Max. Power"}:
        return lambda value: as_int(value)
    if entry in {"Elapsed Time", "Moving Time", "Pace", "Pace (mov.)"}:
        return lambda value: times.to_string(value)
    if entry is None:
        return lambda value: value
    raise ValueError(f"Unknown entry: {entry}")


def split_format(entry: str):
    """Format a value for the splits table."""
    if entry == "Number":
        return lambda value: as_int(value)
    if entry in {"Time", "Split"}:
        return lambda value: times.to_string(value)
    if entry in {"Net Climb", "Ascent"}:
        return lambda value: as_int(value)
    if entry == "Speed":
        return lambda value: f"{value:.2f}"
    raise ValueError(f"Unknown entry: {entry}")


def list_format(entry: str):
    """Format a value for the splits table."""
    if entry in {"Name", "Type", "Server", "User"}:
        return None
    if entry == "Distance":
        return lambda value: f"{value:.2f}"
    if entry == "Start Time":
        return lambda value: str(times.round_time(value))
    raise ValueError(f"Unknown entry: {entry}")


def default_as_string(value) -> str:
    """
    Round a value in a sensible way.

    Always shows at least the nearest integer. Any extra precision is
    limited to the lesser of two decimal places, or three significant
    figures. Also formats timedeltas nicely.
    """
    if isinstance(value, tuple):
        suffix = " " + value[1]
        value = value[0]
    else:
        suffix = ""
    if isinstance(value, timedelta):
        return times.to_string(value)
    if value >= 100:
        return as_int(value) + suffix
    if value >= 10:
        return str(round(value, 1)) + suffix
    return str(round(value, 2)) + suffix



import PyQt5
from PyQt5 import QtWidgets
from PyQt5.QtCore import Qt
from PyQt5.QtGui import QIcon

from activate.app import connect, settings, widgets
from activate.app.dialogs import FormDialog
from activate.app.ui.settings import Ui_settings
from activate.app.widgets import Form


class AddServerDialog(FormDialog):
    def __init__(self, *args, **kwargs):
        layout = {
            "Address": QtWidgets.QLineEdit(),
            "Name": QtWidgets.QLineEdit(),
            "Username": QtWidgets.QLineEdit(),
            "Password": QtWidgets.QLineEdit(),
        }
        layout["Password"].setEchoMode(layout["Password"].Password)
        super().__init__(*args, form=Form(layout), **kwargs)
        self.setWindowTitle("Add Server")


class SettingsDialog(QtWidgets.QDialog, Ui_settings):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.setupUi(self)
        self.settings_tabs.setTabIcon(0, QIcon.fromTheme("measure"))
        self.settings_tabs.setTabIcon(1, QIcon.fromTheme("network-server"))
        self.add_server_button.setIcon(PyQt5.QtGui.QIcon.fromTheme("list-add"))
        self.custom_units = widgets.CustomUnits(self)
        self.units_tab_layout.addLayout(self.custom_units)
        self.units_tab_layout.setAlignment(Qt.AlignTop)

    def add_server(self):
        result = AddServerDialog().exec({})
        if not result:
            return
        self.server_table.add_row()
        self.server_table.set_server(
            self.server_table.rowCount() - 1, connect.Server(*result.values())
        )

    def load_from_settings(self, current_settings: settings.Settings):
        """Load a Settings object to the UI widgets."""
        self.unit_system.setCurrentText(current_settings.unit_system)
        self.server_table.set_servers(current_settings.servers)
        self.custom_units.set_units(current_settings.custom_units)

    def get_settings(self) -> settings.Settings:
        """Get a Settings object from the UI widgets."""
        return settings.Settings(
            unit_system=self.unit_system.currentText(),
            servers=self.server_table.get_servers(),
            custom_units=self.custom_units.units_dict(),
        )

    def exec(self, current_settings, page):
        self.settings_tabs.setCurrentIndex(("Units", "Servers").index(page))
        self.load_from_settings(current_settings)
        result = super().exec()
        if not result:
            return current_settings
        new_settings = self.get_settings()
        new_settings.save()
        return new_settings
